# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
import pytest

import felupe as fe


def pre():

    r = fe.RegionHexahedron(fe.Cube(n=3))

    u = fe.Field(r, dim=3)
    p = fe.Field(r)
    J = fe.Field(r, values=1)

    f = fe.FieldContainer((u, p, J))

    return r, f, (u, p, J)


def test_solve():

    r, _, (u, p, J) = pre()
    f = fe.FieldContainer([u])

    W = fe.constitution.NeoHooke(1, 3)

    F = f.extract()

    bounds = fe.dof.symmetry(u)
    dof0, dof1 = fe.dof.partition(f, bounds)

    ext0 = fe.dof.apply(f, bounds, dof0)

    L = fe.IntegralForm(W.gradient(F)[:-1], f, r.dV)
    a = fe.IntegralForm(W.hessian(F), f, r.dV, f)

    b = L.assemble().toarray()[:, 0]
    A = a.assemble()

    dx = fe.tools.solve(A, b, f, dof0, dof1, f.offsets, ext0)
    assert dx[0].shape == f[0].values.ravel().shape

    fe.tools.save(r, f)
    fe.tools.save(r, f, r=b)
    fe.tools.save(
        r,
        f,
        r=b,
        gradient=W.gradient(F),
    )

    for b in [L.assemble(), L.assemble().toarray(), L.assemble().toarray()[:, 0]]:
        force = fe.tools.force(f, b, bounds["symx"])
        moment = fe.tools.moment(f, b, bounds["symx"])

    for a in [2, 3, 4, 5]:
        curve = fe.tools.curve(np.arange(a), np.ones(a) * force[0])

    s = fe.math.dot(W.gradient(F)[0], fe.math.inv(fe.math.cof(F[0])))

    cauchy = fe.tools.project(fe.math.tovoigt(s), region=r, average=True)
    assert cauchy.shape == (r.mesh.npoints, 6)

    cauchy = fe.tools.project(fe.math.tovoigt(s), region=r, average=False)
    assert cauchy.shape == (r.mesh.cells.size, 6)

    cauchy = fe.tools.project(fe.math.tovoigt(s), region=r, mean=True)
    assert cauchy.shape == (r.mesh.npoints, 6)


def test_solve_mixed():

    r, f, fields = pre()
    u = fields[0]

    f = fe.FieldContainer(fields)

    F, p, J = f.extract()

    W = fe.NeoHooke(1, 3)
    W_mixed = fe.ThreeFieldVariation(W)

    F = u.extract()

    bounds = fe.dof.symmetry(u)
    dof0, dof1 = fe.dof.partition(f, bounds)

    ext0 = fe.dof.apply(f, bounds, dof0)

    L = fe.IntegralForm(W_mixed.gradient([F, p, J])[:-1], f, r.dV)
    a = fe.IntegralForm(W_mixed.hessian([F, p, J]), f, r.dV, f)

    b = L.assemble().toarray()[:, 0]
    A = a.assemble()

    dx = fe.tools.solve(A, b, f, dof0, dof1, f.offsets, ext0)

    assert dx[0].shape == u.values.ravel().shape
    assert dx[1].shape == fields[1].values.ravel().shape
    assert dx[2].shape == fields[2].values.ravel().shape

    fe.tools.save(r, f)
    fe.tools.save(r, f, r=b)
    fe.tools.save(
        r,
        f,
        r=b,
        gradient=W_mixed.gradient([F, p, J]),
    )

    force = fe.tools.force(f, b, bounds["symx"])
    moment = fe.tools.moment(f, b, bounds["symx"])

    for a in [2, 3, 4, 5]:
        curve = fe.tools.curve(np.arange(a), np.ones(a) * force[0])


def test_newton_simple():
    def fun(x):
        return (x - 3) ** 2

    def jac(x):
        return np.array([2 * (x - 3)])

    x0 = np.array([3.1])

    res = fe.tools.newtonrhapson(
        x0, fun, jac, solve=np.linalg.solve, maxiter=32, verbose=True
    )

    assert abs(res.fun) < 1e-6
    assert np.isclose(res.x, 3, rtol=1e-2)

    with pytest.raises(ValueError):
        res = fe.tools.newtonrhapson(
            x0, fun, jac, solve=np.linalg.solve, maxiter=4, verbose=False
        )

    with pytest.raises(ValueError):
        x0 = np.array([np.nan])
        res = fe.tools.newtonrhapson(x0, fun, jac, solve=np.linalg.solve)


def test_newton():

    # create a hexahedron-region on a cube
    region = fe.RegionHexahedron(fe.Cube(n=6))

    # add a displacement field and apply a uniaxial elongation on the cube
    displacement = fe.Field(region, dim=3)
    field = fe.FieldContainer([displacement])
    boundaries, loadcase = fe.dof.uniaxial(field, move=0.2, clamped=True)

    # define the constitutive material behavior
    umat = fe.NeoHooke(mu=1.0, bulk=2.0)

    for kwargs in [{"parallel": True, "umat": umat}]:

        # newton-rhapson procedure
        res = fe.newtonrhapson(
            field,
            verbose=True,
            kwargs=kwargs,
            **loadcase,
        )


def test_newton_plane():

    # create a quad-region on a rectangle
    region = fe.RegionQuad(fe.Rectangle(n=6))

    # add a displacement field and apply a uniaxial elongation on the rectangle
    displacement = fe.Field(region, dim=2)
    field = fe.FieldContainer([displacement])
    boundaries, loadcase = fe.dof.uniaxial(field, move=0.2, clamped=True)

    # define the constitutive material behavior
    umat = fe.LinearElasticPlaneStress(E=1.0, nu=0.3)

    # newton-rhapson procedure
    res = fe.newtonrhapson(
        field,
        verbose=True,
        kwargs=dict(umat=umat),
        **loadcase,
    )

    # define the constitutive material behavior
    umat = fe.LinearElasticPlaneStrain(E=1.0, nu=0.3)

    # newton-rhapson procedure
    res = fe.newtonrhapson(
        field,
        verbose=True,
        kwargs=dict(umat=umat),
        **loadcase,
    )


def test_newton_linearelastic():

    # create a hexahedron-region on a cube
    region = fe.RegionHexahedron(fe.Cube(n=6))

    # add a displacement field and apply a uniaxial elongation on the cube
    displacement = fe.Field(region, dim=3)
    field = fe.FieldContainer([displacement])
    boundaries, loadcase = fe.dof.uniaxial(field, move=0.2, clamped=True)

    # define the constitutive material behavior
    umat = fe.LinearElastic(E=1.0, nu=0.3)

    # newton-rhapson procedure
    res = fe.newtonrhapson(
        field,
        verbose=True,
        kwargs={"umat": umat, "grad": True, "sym": True, "add_identity": False},
        **loadcase,
    )


def test_newton_mixed():

    # create a hexahedron-region on a cube
    mesh = fe.Cube(n=6)
    region = fe.RegionHexahedron(mesh)
    region0 = fe.RegionConstantHexahedron(fe.mesh.convert(mesh, 0))

    # add a displacement field and apply a uniaxial elongation on the cube
    u = fe.Field(region, dim=3)
    p = fe.Field(region0)
    J = fe.Field(region0, values=1)
    field = fe.FieldContainer((u, p, J))

    assert len(field) == 3

    boundaries, loadcase = fe.dof.uniaxial(field, move=0.2, clamped=True)

    # deformation gradient
    F = field.extract(grad=True, sym=False, add_identity=True)

    # define the constitutive material behavior
    nh = fe.NeoHooke(mu=1.0, bulk=2.0)
    umat = fe.ThreeFieldVariation(nh)

    # newton-rhapson procedure
    res = fe.newtonrhapson(x0=field, kwargs=dict(umat=umat), **loadcase)


def test_newton_body():

    # create a hexahedron-region on a cube
    mesh = fe.Cube(n=6)
    region = fe.RegionHexahedron(mesh)
    region0 = fe.RegionConstantHexahedron(fe.mesh.convert(mesh, 0))

    # add a displacement field and apply a uniaxial elongation on the cube
    u = fe.Field(region, dim=3)
    p = fe.Field(region0)
    J = fe.Field(region0, values=1)
    field = fe.FieldContainer((u, p, J))

    boundaries, loadcase = fe.dof.uniaxial(field, move=0.2, clamped=True)

    # define the constitutive material behavior
    nh = fe.NeoHooke(mu=1.0, bulk=2.0)
    umat = fe.ThreeFieldVariation(nh)
    body = fe.SolidBody(umat, field)

    # create a region, a field and a body for a pressure boundary
    regionp = fe.RegionHexahedronBoundary(mesh, only_surface=True)
    fieldp = fe.FieldContainer([fe.Field(regionp, dim=3)])
    bodyp = fe.SolidBodyPressure(fieldp, pressure=1.0)

    # newton-rhapson procedure
    res = fe.newtonrhapson(
        items=[body, bodyp],
        kwargs={},
        **loadcase,
    )


if __name__ == "__main__":
    test_solve()
    test_solve_mixed()
    test_newton_simple()
    test_newton()
    test_newton_mixed()
    test_newton_plane()
    test_newton_linearelastic()
    test_newton_body()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from functools import wraps

import numpy as np
from tensortrax.math import log, sqrt
from tensortrax.math import sum as sum1
from tensortrax.math import trace
from tensortrax.math._linalg import det, eigvalsh, inv
from tensortrax.math._special import from_triu_1d, triu_1d


def isochoric_volumetric_split(fun):
    """Apply the material formulation only on the isochoric part of the
    multiplicative split of the deformation gradient."""

    @wraps(fun)
    def apply_iso(C, *args, **kwargs):
        return fun(det(C) ** (-1 / 3) * C, *args, **kwargs)

    return apply_iso


def saint_venant_kirchhoff(C, mu, lmbda):
    """Strain energy function of the Saint Venant-Kirchhoff material formulation.
    Here, ``I1`` and ``I2`` are strain invariants of the Green-Lagrange strain tensor.
    """
    I1 = trace(C) / 2 - 3 / 2
    I2 = trace(C @ C) / 4 - trace(C) / 2 + 3 / 4
    return mu * I2 + lmbda * I1**2 / 2


def neo_hooke(C, mu):
    "Strain energy function of the Neo-Hookean material formulation."
    return mu / 2 * (det(C) ** (-1 / 3) * trace(C) - 3)


def mooney_rivlin(C, C10, C01):
    "Strain energy function of the Mooney-Rivlin material formulation."
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = J3 * (I1**2 - trace(C @ C)) / 2
    return C10 * (I1 - 3) + C01 * (I2 - 3)


def yeoh(C, C10, C20, C30):
    I1 = det(C) ** (-1 / 3) * trace(C)
    return C10 * (I1 - 3) + C20 * (I1 - 3) ** 2 + C30 * (I1 - 3) ** 3


def third_order_deformation(C, C10, C01, C11, C20, C30):
    "Strain energy function of the Third-Order-Deformation material formulation."
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = J3 * (I1**2 - trace(C @ C)) / 2
    return (
        C10 * (I1 - 3)
        + C01 * (I2 - 3)
        + C11 * (I1 - 3) * (I2 - 3)
        + C20 * (I1 - 3) ** 2
        + C30 * (I1 - 3) ** 3
    )


def ogden(C, mu, alpha):
    "Strain energy function of the Ogden material formulation."
    wC = det(C) ** (-1 / 3) * eigvalsh(C)
    return sum1([2 * m / a**2 * (sum1(wC ** (a / 2)) - 3) for m, a in zip(mu, alpha)])


def arruda_boyce(C, C1, limit):
    "Strain energy function of the Arruda-Boyce material formulation."
    I1 = det(C) ** (-1 / 3) * trace(C)

    alpha = [1 / 2, 1 / 20, 11 / 1050, 19 / 7000, 519 / 673750]
    beta = 1 / limit**2

    out = []
    for i, a in enumerate(alpha):
        j = i + 1
        out.append(a * beta ** (2 * j - 2) * (I1**j - 3**j))

    return C1 * sum1(out)


def extended_tube(C, Gc, delta, Ge, beta):
    "Strain energy function of the Extended-Tube material formulation."
    J3 = det(C) ** (-1 / 3)
    D = J3 * trace(C)
    wC = J3 * eigvalsh(C)
    g = (1 - delta**2) * (D - 3) / (1 - delta**2 * (D - 3))
    Wc = Gc / 2 * (g + log(1 - delta**2 * (D - 3)))
    We = 2 * Ge / beta**2 * sum1(wC ** (-beta / 2) - 1)
    return Wc + We


def van_der_waals(C, mu, limit, a, beta):
    "Strain energy function of the Van der Waals material formulation."
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = J3 * (trace(C) ** 2 - trace(C @ C)) / 2
    Im = (1 - beta) * I1 + beta * I2
    Im.x[np.isclose(Im.x, 3)] += 1e-8
    eta = sqrt((Im - 3) / (limit**2 - 3))
    return mu * (
        -(limit**2 - 3) * (log(1 - eta) + eta) - 2 / 3 * a * ((Im - 3) / 2) ** (3 / 2)
    )


@isochoric_volumetric_split
def finite_strain_viscoelastic(C, Cin, mu, eta, dtime):
    "Finite strain viscoelastic material formulation."

    # update of state variables by evolution equation
    Ci = from_triu_1d(Cin, like=C) + mu / eta * dtime * C
    Ci = det(Ci) ** (-1 / 3) * Ci

    # first invariant of elastic part of right Cauchy-Green deformation tensor
    I1 = trace(C @ inv(Ci))

    # strain energy function and state variable
    return mu / 2 * (I1 - 3), triu_1d(Ci)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

try:
    from einsumt import einsumt
except ModuleNotFoundError:
    from numpy import einsum as einsumt


class Basis:
    r"""A basis and its gradient built on top of a scalar- or vector-valued
    field. *Basis* refers to the trial and test field, either values or
    gradients evaluated at quadrature points. The first two indices of a basis
    are used for looping over the element shape functions ``a`` and its
    components ``i``. The third index represents the vector component ``j`` of
    the field. The two trailing axes ``(p, c)`` contain the evaluated element
    shape functions at quadrature points ``p`` per cell ``c``.

    ..  math::

        \varphi_{aijpc} = \delta_{ij} \left( h_a \right)_{pc}


    For gradients, the fourth index is used for the vector component of the
    partial derivative ``k``.

    ..  math::

        \text{grad}(\varphi)_{aijkpc} = \delta_{ij}
            \left( \frac{\partial h_a}{\partial X_K} \right)_{pc}

    Parameters
    ----------
    field : Field
        A field on which the basis should be created.
    parallel : bool, optional (default is False)
        Flag to activate parallel (threaded) basis evaluation.

    Attributes
    ----------
    basis : ndarray
        The evaluated basis functions at quadrature points.
    grad : ndarray
        The evaluated gradient of the basis (if provided by the region).

    """

    def __init__(self, field, parallel=False):

        self.field = field

        einsum = einsumt if parallel else np.einsum

        self.basis = einsum(
            "ij,apc->aijpc",
            np.eye(self.field.dim),
            self.field.region.h,
        )

        if hasattr(self.field.region, "dhdX"):
            self.grad = einsum(
                "ij,akpc->aijkpc", np.eye(self.field.dim), self.field.region.dhdX
            )

        else:
            self.grad = None


class BasisMixed:
    r"""A basis and its gradient built on top of a scalar- or vector-valued
    field container. *Basis* refers to the trial and test field, either values or
    gradients evaluated at quadrature points. The first two indices of a basis
    are used for looping over the element shape functions ``a`` and its
    components ``i``. The third index represents the vector component ``j`` of
    the field. The two trailing axes ``(p, c)`` contain the evaluated element
    shape functions at quadrature points ``p`` per cell ``c``.

    ..  math::

        \varphi_{aijpc} = \delta_{ij} \left( h_a \right)_{pc}


    For gradients, the fourth index is used for the vector component of the
    partial derivative ``k``.

    ..  math::

        \text{grad}(\varphi)_{aijkpc} = \delta_{ij}
            \left( \frac{\partial h_a}{\partial X_K} \right)_{pc}

    Parameters
    ----------
    field : FieldContainer
        A field container on which the basis should be created.

    Attributes
    ----------
    basis : ndarray
        The evaluated basis functions at quadrature points.
    grad : ndarray
        The evaluated gradient of the basis (if provided by the region).

    """

    def __init__(self, field, parallel=False):

        self.field = field
        self.basis = [Basis(f, parallel=parallel) for f in self.field]

    def __getitem__(self, idx):
        "Slice-based access to underlying bases."

        return self.basis[idx]



from ._container import MeshContainer
from ._convert import (
    add_midpoints_edges,
    add_midpoints_faces,
    add_midpoints_volumes,
    collect_edges,
    collect_faces,
    collect_volumes,
    convert,
)
from ._dual import dual
from ._geometry import (
    Cube,
    CubeArbitraryOrderHexahedron,
    Grid,
    Line,
    Rectangle,
    RectangleArbitraryOrderQuad,
)
from ._line_rectangle_cube import cube_hexa as _cube_hexa
from ._line_rectangle_cube import line_line as _line_line
from ._line_rectangle_cube import rectangle_quad as _rectangle_quad
from ._mesh import Mesh
from ._read import read
from ._tools import (
    concatenate,
    expand,
    flip,
    mirror,
    revolve,
    rotate,
    runouts,
    stack,
    sweep,
    translate,
    triangulate,
)

__all__ = [
    "_cube_hexa",
    "_line_line",
    "_rectangle_quad",
    "MeshContainer",
    "add_midpoints_edges",
    "add_midpoints_faces",
    "add_midpoints_volumes",
    "collect_edges",
    "collect_faces",
    "collect_volumes",
    "convert",
    "dual",
    "Cube",
    "CubeArbitraryOrderHexahedron",
    "Grid",
    "Line",
    "Rectangle",
    "RectangleArbitraryOrderQuad",
    "Mesh",
    "read",
    "concatenate",
    "expand",
    "flip",
    "mirror",
    "revolve",
    "rotate",
    "runouts",
    "stack",
    "sweep",
    "translate",
    "triangulate",
]



from ._boundary import Boundary
from ._loadcase import biaxial, planar, shear, symmetry, uniaxial
from ._tools import apply, get_dof0, get_dof1, partition

__all__ = [
    "Boundary",
    "biaxial",
    "planar",
    "shear",
    "symmetry",
    "uniaxial",
    "apply",
    "get_dof0",
    "get_dof1",
    "partition",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.sparse import csr_matrix as sparsematrix

from .._field import Field
from ..region import Region


def topoints(values, region, sym=True, mode="tensor"):

    rows = region.mesh.cells.T.ravel()
    cols = np.zeros_like(rows)

    if mode == "tensor":
        dim = values.shape[0]
        if sym:
            if dim == 3:
                ij = [(0, 0), (1, 1), (2, 2), (0, 1), (1, 2), (0, 2)]
            elif dim == 2:
                ij = [(0, 0), (1, 1), (0, 1)]
        else:
            if dim == 3:
                ij = [
                    (0, 0),
                    (0, 1),
                    (0, 2),
                    (1, 0),
                    (1, 1),
                    (1, 2),
                    (2, 0),
                    (2, 1),
                    (2, 2),
                ]
            elif dim == 2:
                ij = [(0, 0), (0, 1), (1, 0), (1, 1)]

        out = Field(region, dim=len(ij)).values

        for a, (i, j) in enumerate(ij):
            out[:, a] = (
                sparsematrix(
                    (values.reshape(dim, dim, -1)[i, j], (rows, cols)),
                    shape=(region.mesh.npoints, 1),
                ).toarray()[:, 0]
                / region.mesh.cells_per_point
            )

    elif mode == "scalar":
        out = sparsematrix(
            (values.ravel(), (rows, cols)), shape=(region.mesh.npoints, 1)
        ).toarray()[:, 0]
        out = out / region.mesh.cells_per_point

    return out


def project(values, region, average=True, mean=False):
    """Projection (and optionally averaging) of scalar or vectorial values
    at quadrature points to mesh-points.
    """

    dim = values.shape[:-2]
    size = int(np.product(dim))
    weights = region.quadrature.weights

    # transpose values
    idx = np.arange(len(values.shape))
    idx[: len(dim)] = idx[: len(dim)][::-1]
    values = values.transpose(idx)

    if mean:

        # evaluate how often the values must be repeated to match the number
        # of element-points
        reps = np.ones(len(values.shape), dtype=int)
        reps[-2] = len(region.element.points)

        # np.average(keepdims=True) requires numpy >= 1.23.0
        values = np.average(values, axis=-2, weights=weights)

        # workaround for np.average(keepdims=True)
        shape = values.shape
        shape = np.insert(shape, -1, 1)
        values = values.reshape(*shape)

        # broadcast averaged values to match the number of element-points
        values = np.broadcast_to(values, shape=np.broadcast_shapes(shape, reps))

    # reshape from (shape, nint.points, nelements) to 1d-values
    u = values.T.reshape(-1, size)

    # disconnect the mesh
    m = region.mesh.disconnect()

    if mean:
        # region on disconnected mesh with original quadrature scheme
        # a single-point quadrature would be sufficient
        # but would require additional (not available) informations
        r = Region(m, region.element, region.quadrature, grad=False)
    else:
        # region on disconnected mesh with inverse quadrature scheme
        r = Region(m, region.element, region.quadrature.inv(), grad=False)

    # field for values on disconnected mesh; project values to mesh-points
    f = Field(r, dim=size, values=u)
    v = f.interpolate()

    if mean:

        # due to the usage of the original quadrature scheme the averaging must be
        # applied again
        # np.average(keepdims=True) requires numpy >= 1.23.0
        v = np.average(v, axis=-2, weights=weights).reshape(size, 1, -1)

        # broadcast averaged values to match the number of element-points
        shape = np.array([*v.shape[:-2], len(region.element.points), v.shape[-1]])
        v = np.broadcast_to(v, shape=shape)

    if average:

        # create dummy field for values on original mesh
        # (used for calculation of sparse-matrix indices)
        g = Field(region, dim=size)

        # average values
        w = sparsematrix(
            (v.T.ravel(), g.indices.ai), shape=(size * region.mesh.npoints, 1)
        ).toarray().reshape(-1, size) / region.mesh.cells_per_point.reshape(-1, 1)

    else:

        w = v.T.reshape(-1, size)

    return w



from ._misc import logo, runs_on
from ._newton import fun_items as fun
from ._newton import jac_items as jac
from ._newton import newtonrhapson
from ._plot import ViewField, ViewMesh, ViewXdmf
from ._post import curve, force, moment
from ._project import project, topoints
from ._save import save
from ._solve import solve

__all__ = [
    "fun",
    "jac",
    "newtonrhapson",
    "curve",
    "force",
    "moment",
    "project",
    "topoints",
    "logo",
    "runs_on",
    "save",
    "solve",
    "ViewMesh",
    "ViewField",
    "ViewXdmf",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.sparse import eye, lil_matrix

from ._helpers import Assemble, Results


class MultiPointConstraint:
    def __init__(
        self, field, points, centerpoint, skip=(False, False, False), multiplier=1e3
    ):
        "RBE2 Multi-point-constraint."
        self.field = field
        self.mesh = field.region.mesh
        self.points = np.asarray(points)
        self.centerpoint = centerpoint
        self.mask = ~np.array(skip, dtype=bool)[: self.mesh.dim]
        self.axes = np.arange(self.mesh.dim)[self.mask]
        self.multiplier = multiplier

        self.results = Results(stress=False, elasticity=False)
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)

    def _vector(self, field=None, parallel=False):
        "Calculate vector of residuals with RBE2 contributions."

        if field is not None:
            self.field = field

        u = self.field.fields[0].values
        N = self.multiplier * (-u[self.points] + u[self.centerpoint])
        N[:, ~self.mask] = 0

        r = lil_matrix(u.shape)
        r[self.points] = -N
        r[self.centerpoint] = N.sum(axis=0)

        self.results.force = r.reshape(-1, 1).tocsr()
        return self.results.force

    def _matrix(self, field=None, parallel=False):
        "Calculate stiffness with RBE2 contributions."

        if field is not None:
            self.field = field

        indices = np.arange(self.mesh.ndof).reshape(self.mesh.points.shape)
        td = [indices[self.points.reshape(-1, 1), ax].ravel() for ax in self.axes]
        cd = [indices[self.centerpoint, ax].ravel() for ax in self.axes]

        L = lil_matrix((self.mesh.ndof, self.mesh.ndof))

        for t, c in zip(td, cd):

            L[t.reshape(-1, 1), t] = eye(len(t)) * self.multiplier
            L[t.reshape(-1, 1), c] = -self.multiplier
            L[c.reshape(-1, 1), t] = -self.multiplier
            L[c.reshape(-1, 1), c] = eye(len(c)) * self.multiplier * len(self.points)

        self.results.stiffness = L.tocsr()
        return self.results.stiffness


class MultiPointContact:
    def __init__(
        self, field, points, centerpoint, skip=(False, False, False), multiplier=1e6
    ):
        "RBE2 Multi-point-bolt-constraint."
        self.field = field
        self.mesh = field.region.mesh
        self.points = np.asarray(points)
        self.centerpoint = centerpoint
        self.mask = ~np.array(skip, dtype=bool)[: self.mesh.dim]
        self.axes = np.arange(self.mesh.dim)[self.mask]
        self.multiplier = multiplier

        self.results = Results(stress=False, elasticity=False)
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)

    def _vector(self, field=None, parallel=False):
        "Calculate vector of residuals with RBE2 contributions."

        if field is not None:
            self.field = field

        u = self.field.fields[0].values

        Xc = self.mesh.points[self.centerpoint]
        Xt = self.mesh.points[self.points]

        xc = u[self.centerpoint] + Xc
        xt = u[self.points] + Xt

        mask = np.sign(-Xt + Xc) == np.sign(-xt + xc)
        mask[:, ~self.mask] = True
        n = -xt + xc
        n[mask] = 0

        r = lil_matrix(u.shape)
        r[self.points] = -self.multiplier * n
        r[self.centerpoint] = self.multiplier * n.sum(axis=0)

        self.results.force = r.reshape(-1, 1).tocsr()
        return self.results.force

    def _matrix(self, field=None, parallel=False):
        "Calculate stiffness with RBE2 contributions."

        if field is not None:
            self.field = field

        u = self.field.fields[0].values

        Xc = self.mesh.points[self.centerpoint]
        Xt = self.mesh.points[self.points]

        xc = u[self.centerpoint] + Xc
        xt = u[self.points] + Xt

        mask = np.sign(-Xt + Xc) != np.sign(-xt + xc)
        masks = [mask[:, ax] for ax in self.axes]

        indices = np.arange(self.mesh.ndof).reshape(self.mesh.points.shape)
        td = [indices[self.points.reshape(-1, 1), ax].ravel() for ax in self.axes]
        cd = [indices[self.centerpoint, ax].ravel() for ax in self.axes]

        L = lil_matrix((self.mesh.ndof, self.mesh.ndof))

        for t, c, m in zip(td, cd, masks):

            L[t[m].reshape(-1, 1), t[m]] = eye(len(t[m])) * self.multiplier
            L[t[m].reshape(-1, 1), c] = -self.multiplier
            L[c.reshape(-1, 1), t[m]] = -self.multiplier
            L[c.reshape(-1, 1), c] = eye(len(c)) * self.multiplier * len(self.points[m])

        self.results.stiffness = L.tocsr()
        return self.results.stiffness



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from .._assembly import IntegralFormMixed
from ..constitution import AreaChange
from ._helpers import Assemble, Results


class SolidBodyPressure:
    "A hydrostatic pressure boundary on a SolidBody."

    def __init__(self, field, pressure=None):

        self.field = field
        self._normals = self.field.region.normals

        self.results = Results()
        self.results.kinematics = self._extract(self.field)

        if pressure is not None:
            self.results.pressure = pressure
        else:
            self.results.pressure = 0

        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)
        self._area_change = AreaChange()

    def update(self, pressure):

        self.__init__(self.field, pressure)

    def _extract(self, field):

        self.field = field
        self.results.kinematics = self.field.extract()

        return self.results.kinematics

    def _vector(self, field=None, pressure=None, parallel=False, resize=None):

        if field is not None:

            self._update(field)
            self.results.kinematics = self._extract(self.field)

        fun = self._area_change.function(
            self.results.kinematics,
            self._normals,
            parallel=parallel,
        )

        if pressure is not None:
            self.results.pressure = pressure

        fun[0] *= self.results.pressure

        self.results.force = IntegralFormMixed(
            fun=fun, v=self.field, dV=self.field.region.dV, grad_v=[False]
        ).assemble(parallel=parallel)

        if resize is not None:
            self.results.force.resize(*resize.shape)

        return self.results.force

    def _matrix(self, field=None, pressure=None, parallel=False, resize=None):

        if field is not None:

            self._update(field)
            self.results.kinematics = self._extract(self.field)

        fun = self._area_change.gradient(
            self.results.kinematics,
            self._normals,
            parallel=parallel,
        )

        if pressure is not None:
            self.results.pressure = pressure

        fun[0] *= self.results.pressure

        self.results.stiffness = IntegralFormMixed(
            fun=fun,
            v=self.field,
            u=self.field,
            dV=self.field.region.dV,
            grad_v=[False],
            grad_u=[True],
        ).assemble(parallel=parallel)

        if resize is not None:
            self.results.stiffness.resize(*resize.shape)

        return self.results.stiffness

    def _update(self, other_field, field=None):

        if field is not None:
            self.field = field

        self.field[0].values = other_field[0].values
        self.results.kinematics = self._extract(self.field)

        return self.field



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

try:
    from einsumt import einsumt
except ModuleNotFoundError:
    from numpy import einsum as einsumt


def identity(A=None, dim=None, shape=None):
    "Identity according to matrix A with optional specified dim."
    if A is not None:
        dimA = A.shape[0]
        shapeA = A.shape[2:]
        if dim is None:
            dim = dimA
        if shape is None:
            shape = shapeA

    ones = (1,) * len(shape)
    eye = np.eye(dim).reshape(dim, dim, *ones)

    return np.ascontiguousarray(np.broadcast_to(eye, (dim, dim, *shape)))


def sym(A):
    "Symmetric part of matrix A."
    return (A + transpose(A)) / 2


def dya(A, B, mode=2, parallel=False):
    "Dyadic (outer or kronecker) product of two tensors."

    if parallel:
        einsum = einsumt
    else:
        einsum = np.einsum

    if mode == 2:
        return einsum("ij...,kl...->ijkl...", A, B)
    elif mode == 1:
        return einsum("i...,j...->ij...", A, B)
    else:
        raise ValueError("unknown mode. (1 or 2)", mode)


def inv(A, determinant=None, full_output=False, sym=False):
    """ "Inverse of A with optionally pre-calculated determinant,
    optional additional output of the calculated determinant or
    a simplified calculation of the inverse for sym. matrices."""

    detAinvA = np.zeros_like(A)

    if determinant is None:
        detA = det(A)
    else:
        detA = determinant

    if A.shape[0] == 3:

        detAinvA[0, 0] = -A[1, 2] * A[2, 1] + A[1, 1] * A[2, 2]
        detAinvA[1, 1] = -A[0, 2] * A[2, 0] + A[0, 0] * A[2, 2]
        detAinvA[2, 2] = -A[0, 1] * A[1, 0] + A[0, 0] * A[1, 1]

        detAinvA[0, 1] = A[0, 2] * A[2, 1] - A[0, 1] * A[2, 2]
        detAinvA[0, 2] = -A[0, 2] * A[1, 1] + A[0, 1] * A[1, 2]
        detAinvA[1, 2] = A[0, 2] * A[1, 0] - A[0, 0] * A[1, 2]

        if sym:
            detAinvA[1, 0] = detAinvA[0, 1]
            detAinvA[2, 0] = detAinvA[0, 2]
            detAinvA[2, 1] = detAinvA[1, 2]
        else:
            detAinvA[1, 0] = A[1, 2] * A[2, 0] - A[1, 0] * A[2, 2]
            detAinvA[2, 0] = -A[1, 1] * A[2, 0] + A[1, 0] * A[2, 1]
            detAinvA[2, 1] = A[0, 1] * A[2, 0] - A[0, 0] * A[2, 1]

    elif A.shape[0] == 2:
        detAinvA[0, 0] = A[1, 1]
        detAinvA[0, 1] = -A[0, 1]
        detAinvA[1, 0] = -A[1, 0]
        detAinvA[1, 1] = A[0, 0]

    if full_output:
        return detAinvA / detA, detA
    else:
        return detAinvA / detA


def det(A):
    "Determinant of matrix A."
    if A.shape[0] == 3:
        detA = (
            A[0, 0] * A[1, 1] * A[2, 2]
            + A[0, 1] * A[1, 2] * A[2, 0]
            + A[0, 2] * A[1, 0] * A[2, 1]
            - A[2, 0] * A[1, 1] * A[0, 2]
            - A[2, 1] * A[1, 2] * A[0, 0]
            - A[2, 2] * A[1, 0] * A[0, 1]
        )
    elif A.shape[0] == 2:
        detA = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    elif A.shape[0] == 1:
        detA = A[0, 0]
    return detA


def dev(A):
    "Deviator of matrix A."
    dim = A.shape[0]
    return A - trace(A) / dim * identity(A)


def cof(A):
    "Cofactor matrix of A (as a wrapper for the inverse of A)."
    return transpose(inv(A, determinant=1.0))


def eig(A, eig=np.linalg.eig):
    "Eigenvalues and -vectors of matrix A."
    wA, vA = eig(A.T)
    return wA.T, vA.T


def eigh(A):
    "Eigenvalues and -vectors of a symmetric matrix A."
    return eig(A, eig=np.linalg.eigh)


def eigvals(A, shear=False, eig=np.linalg.eig):
    "Eigenvalues (and optional principal shear values) of a matrix A."
    wA = eig(A.T)[0].T
    if shear:
        dim = wA.shape[0]
        if dim == 3:
            ij = [(1, 0), (2, 0), (2, 1)]
        elif dim == 2:
            ij = [(1, 0)]
        dwA = np.array([wA[i] - wA[j] for i, j in ij])
        return np.vstack((wA, dwA))
    else:
        return wA


def eigvalsh(A, shear=False):
    "Eigenvalues (and optional principal shear values) of a symmetric matrix A."
    return eigvals(A, shear=shear, eig=np.linalg.eigh)


def transpose(A, mode=1):
    "Transpose (mode=1) or major-transpose (mode=2) of matrix A."
    if mode == 1:
        return np.einsum("ij...->ji...", A)
    elif mode == 2:
        return np.einsum("ijkl...->klij...", A)
    else:
        raise ValueError("Unknown value of mode.")


def majortranspose(A):
    return transpose(A, mode=2)


def trace(A):
    "The sum of the diagonal elements of A."
    return np.trace(A)


def cdya_ik(A, B, parallel=False):
    "ik - crossed dyadic-product of A and B."
    if parallel:
        einsum = einsumt
    else:
        einsum = np.einsum
    return einsum("ij...,kl...->ikjl...", A, B)


def cdya_il(A, B, parallel=False):
    "il - crossed dyadic-product of A and B."
    if parallel:
        einsum = einsumt
    else:
        einsum = np.einsum
    return einsum("ij...,kl...->ilkj...", A, B)


def cdya(A, B, parallel=False):
    "symmetric - crossed dyadic-product of A and B."
    return (cdya_ik(A, B, parallel=parallel) + cdya_il(A, B, parallel=parallel)) * 0.5


def cross(a, b):
    "Cross product of two vectors a and b."
    return np.einsum(
        "...i->i...", np.cross(np.einsum("i...->...i", a), np.einsum("i...->...i", b))
    )


def dot(A, B, mode=(2, 2), parallel=False):
    "Dot-product of A and B with inputs of n trailing axes.."

    if parallel:
        einsum = einsumt
    else:
        einsum = np.einsum

    if mode == (2, 2):
        return einsum("ik...,kj...->ij...", A, B)

    elif mode == (1, 1):
        return einsum("i...,i...->...", A, B)

    elif mode == (4, 4):
        return einsum("ijkp...,plmn...->ijklmn...", A, B)

    elif mode == (2, 1):
        return einsum("ij...,j...->i...", A, B)

    elif mode == (1, 2):
        return einsum("i...,ij...->j...", A, B)

    elif mode == (4, 1):
        return einsum("ijkl...,l...->ijk...", A, B)

    elif mode == (1, 4):
        return einsum("i...,ijkl...->jkl...", A, B)

    elif mode == (2, 4):
        return einsum("im...,mjkl...->ijkl...", A, B)

    elif mode == (4, 2):
        return einsum("ijkm...,ml...->ijkl...", A, B)

    else:
        raise TypeError("Unknown shape of A and B.")


def ddot(A, B, mode=(2, 2), parallel=False):
    "Double-Dot-product of A and B with inputs of `n` trailing axes."

    if parallel:
        einsum = einsumt
    else:
        einsum = np.einsum

    if mode == (2, 2):
        return einsum("ij...,ij...->...", A, B)
    elif mode == (2, 4):
        return einsum("ij...,ijkl...->kl...", A, B)
    elif mode == (4, 2):
        return einsum("ijkl...,kl...->ij...", A, B)
    elif mode == (4, 4):
        return einsum("ijkl...,klmn...->ijmn...", A, B)
    else:
        raise TypeError("Unknown shape of A and B.")


def tovoigt(A, strain=False):
    "Convert (3, 3) tensor to (6, ) voigt notation."
    dim = A.shape[0]
    if dim == 2:
        B = np.zeros((3, *A.shape[2:]))
        ij = [(0, 0), (1, 1), (0, 1)]
    else:
        B = np.zeros((6, *A.shape[2:]))
        ij = [(0, 0), (1, 1), (2, 2), (0, 1), (1, 2), (0, 2)]
    for i6, (i, j) in enumerate(ij):
        B[i6] = A[i, j]
    if strain:
        B[dim:] *= 2
    return B


def reshape(A, shape, trailing_axes=2):
    return A.reshape(np.append(shape, A.shape[-trailing_axes:]))


def ravel(A, trailing_axes=2):
    ij, shape = np.split(A.shape, [-trailing_axes])
    return reshape(A, shape=np.product(ij))



from numpy import sqrt

from ._field import defgrad, extract, grad, interpolate, norm, strain, values
from ._math import linsteps
from ._spatial import rotation_matrix
from ._tensor import (
    cdya,
    cdya_ik,
    cdya_il,
    cof,
    cross,
    ddot,
    det,
    dev,
    dot,
    dya,
    eig,
    eigh,
    eigvals,
    eigvalsh,
    identity,
    inv,
    majortranspose,
    ravel,
    reshape,
    sym,
    tovoigt,
    trace,
    transpose,
)

__all__ = [
    "sqrt",
    "defgrad",
    "extract",
    "grad",
    "interpolate",
    "norm",
    "strain",
    "values",
    "linsteps",
    "rotation_matrix",
    "cdya",
    "cdya_ik",
    "cdya_il",
    "cof",
    "cross",
    "ddot",
    "det",
    "dev",
    "dot",
    "dya",
    "eig",
    "eigh",
    "eigvals",
    "eigvalsh",
    "identity",
    "inv",
    "majortranspose",
    "ravel",
    "reshape",
    "sym",
    "tovoigt",
    "trace",
    "transpose",
]



from ._kinematics import AreaChange, LineChange, VolumeChange
from ._mixed import ThreeFieldVariation
from ._models_hyperelasticity import NeoHooke
from ._models_hyperelasticity_ad import (
    arruda_boyce,
    extended_tube,
    finite_strain_viscoelastic,
    isochoric_volumetric_split,
    mooney_rivlin,
    neo_hooke,
    ogden,
    saint_venant_kirchhoff,
    third_order_deformation,
    van_der_waals,
    yeoh,
)
from ._models_linear_elasticity import (
    LinearElastic,
    LinearElasticPlaneStrain,
    LinearElasticPlaneStress,
    LinearElasticTensorNotation,
    lame_converter,
)
from ._models_pseudo_elasticity import OgdenRoxburgh
from ._user_materials import (
    LinearElasticPlasticIsotropicHardening,
    UserMaterial,
    UserMaterialStrain,
)
from ._user_materials_hyperelastic import UserMaterialHyperelastic
from ._user_materials_models import (
    linear_elastic,
    linear_elastic_plastic_isotropic_hardening,
)

__all__ = [
    "NeoHooke",
    "arruda_boyce",
    "extended_tube",
    "finite_strain_viscoelastic",
    "isochoric_volumetric_split",
    "mooney_rivlin",
    "neo_hooke",
    "ogden",
    "saint_venant_kirchhoff",
    "third_order_deformation",
    "van_der_waals",
    "yeoh",
    "LinearElastic",
    "LinearElasticPlaneStrain",
    "LinearElasticPlaneStress",
    "LinearElasticTensorNotation",
    "lame_converter",
    "OgdenRoxburgh",
    "LinearElasticPlasticIsotropicHardening",
    "UserMaterial",
    "UserMaterialStrain",
    "UserMaterialHyperelastic",
    "AreaChange",
    "LineChange",
    "VolumeChange",
    "ThreeFieldVariation",
    "linear_elastic",
    "linear_elastic_plastic_isotropic_hardening",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy

import numpy as np

from ..math import identity
from ..math import sym as symmetric
from ._indices import Indices


class Field:
    r"""A Field on points of a `region` with dimension `dim`
    and initial point `values`. A slice of this field directly
    accesses the field values as 1d-array.

    The interpolation method returns the field values evaluated at the
    numeric integration points ``p`` of all cells ``c`` in the region.

    ..  math::

        u^i_{(pc)} = \hat{u}_a^i h_{a(pc)}

    The gradient method returns the gradient of the field values w.r.t. the
    undeformed mesh point coordinates, evaluated at the integration points of
    all cells in the region.

    ..  math::

        \left( \frac{\partial u^i}{\partial X^J} \right)_{(pc)} =
        \hat{u}^i_{a(pc)}
        \left( \frac{\partial h_a}{\partial X^J} \right)_{(pc)}

    Arguments
    ---------
    region : Region
        The region on which the field will be created.
    dim : int (default is 1)
        The dimension of the field.
    values : float (default is 0.0) or array
        A single value for all components of the field or an array of
        shape (region.mesh.npoints, dim)`.
    kwargs : dict, optional
        Optional keyword arguments of the field.

    """

    def __init__(self, region, dim=1, values=0, **kwargs):

        self.region = region
        self.dim = dim
        self.shape = self.region.quadrature.npoints, self.region.mesh.ncells

        # set optional user-defined attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        # init values
        if isinstance(values, np.ndarray):
            if values.size == dim:
                shape = (region.mesh.npoints, dim)
                self.values = np.broadcast_to(values.reshape(-1, dim), shape)
            elif len(values) == region.mesh.npoints:
                self.values = values
            else:
                raise ValueError("Wrong shape of values.")

        else:  # scalar value
            self.values = np.ones((region.mesh.npoints, dim)) * values

        eai, ai = self._indices_per_cell(self.region.mesh.cells, dim)
        self.indices = Indices(eai, ai, region, dim)

    def _indices_per_cell(self, cells, dim):
        "Calculate pre-defined indices for sparse matrices."

        # index of cell "e", point "a" and component "i"
        eai = (
            dim * np.repeat(cells, dim) + np.tile(np.arange(dim), cells.size)
        ).reshape(*cells.shape, dim)
        # store indices as (rows, cols) (note: sparse-matrices are always 2d)
        ai = (eai.ravel(), np.zeros_like(eai.ravel()))

        return eai, ai

    def grad(self, sym=False):
        """Gradient as partial derivative of field values at points w.r.t.
        undeformed coordinates, evaluated at the integration points of
        all cells in the region. Optionally, the symmetric part of
        the gradient is evaluated.

        Arguments
        ---------
        sym : bool, optional (default is False)
            Calculate the symmetric part of the gradient.

        Returns
        -------
        array
            Gradient as partial derivative of field values at points w.r.t.
            undeformed coordinates, evaluated at the integration points of
            all cells in the region.
        """

        # gradient dudX_IJpe as partial derivative of field values at points "aI"
        # w.r.t. undeformed coordinates "J" evaluated at quadrature point "p"
        # for each cell "c"
        g = np.einsum(
            "ca...,aJpc->...Jpc",
            self.values[self.region.mesh.cells],
            self.region.dhdX,
        )

        if sym:
            return symmetric(g)
        else:
            return g

    def interpolate(self):
        """Interpolate field values at points and evaluate them at the
        integration points of all cells in the region."""

        # interpolated field values "aI"
        # evaluated at quadrature point "p"
        # for cell "c"
        return np.einsum(
            "ca...,apc->...pc", self.values[self.region.mesh.cells], self.region.h
        )

    def extract(self, grad=True, sym=False, add_identity=True):
        """Generalized extraction method which evaluates either the gradient
        or the field values at the integration points of all cells
        in the region. Optionally, the symmetric part of the gradient is
        evaluated and/or the identity matrix is added to the gradient.

        Arguments
        ---------
        grad : bool, optional (default is True)
            Flag for gradient evaluation.
        sym : bool, optional (default is False)
            Flag for symmetric part if the gradient is evaluated.
        add_identity : bool, optional (default is True)
            Flag for the addition of the identity matrix
            if the gradient is evaluated.

        Returns
        -------
        array
            (Symmetric) gradient or interpolated field values evaluated at
            the integration points of each cell in the region.
        """

        if grad:
            gr = self.grad()

            if sym:
                gr = symmetric(gr)

            if add_identity:
                gr = identity(gr) + gr

            return gr
        else:
            return self.interpolate()

    def copy(self):
        "Return a copy of the field."
        return deepcopy(self)

    def fill(self, a):
        "Fill all field values with a scalar value."
        self.values.fill(a)

    def __add__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            field = deepcopy(self)
            field.values += newvalues.reshape(-1, field.dim)
            return field

        elif isinstance(newvalues, Field):
            field = deepcopy(self)
            field.values += newvalues.values
            return field

        else:
            raise TypeError("Unknown type.")

    def __sub__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            field = deepcopy(self)
            field.values -= newvalues.reshape(-1, field.dim)
            return field

        elif isinstance(newvalues, Field):
            field = deepcopy(self)
            field.values -= newvalues.values
            return field

        else:
            raise TypeError("Unknown type.")

    def __mul__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            field = deepcopy(self)
            field.values *= newvalues.reshape(-1, field.dim)
            return field

        elif isinstance(newvalues, Field):
            field = deepcopy(self)
            field.values *= newvalues.values
            return field

        else:
            raise TypeError("Unknown type.")

    def __truediv__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            field = deepcopy(self)
            field.values /= newvalues.reshape(-1, field.dim)
            return field

        elif isinstance(newvalues, Field):
            field = deepcopy(self)
            field.values /= newvalues.values
            return field

        else:
            raise TypeError("Unknown type.")

    def __iadd__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            self.values += newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values += newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __isub__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            self.values -= newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values -= newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __imul__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            self.values *= newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values *= newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __itruediv__(self, newvalues):

        if isinstance(newvalues, np.ndarray):
            self.values /= newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values /= newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __getitem__(self, dof):
        """Slice-based access to a 1d-representation of field values by
        a list of degrees of freedom `dof`."""

        return self.values.ravel()[dof]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import tensortrax as tr

from ..math import cdya_ik, dot, transpose
from ._user_materials import UserMaterial


class UserMaterialHyperelastic(UserMaterial):
    """A user-defined hyperelastic material definition with a given function
    for the strain energy function with Automatic Differentiation provided by
    ``tensortrax``.

    Take this code-block as template

    ..  code-block::

        import tensortrax.math as tm

        def neo_hooke(C, mu):
            "Strain energy function of the Neo-Hookean material formulation."
            return mu / 2 * (tm.linalg.det(C) ** (-1/3) * tm.trace(C) - 3)

        umat = fem.UserMaterialHyperelastic(neo_hooke, mu=1)

    and this code-block for material formulations with state variables:

    ..  code-block::

        import tensortrax.math as tm

        def viscoelastic(C, Cin, mu, eta, dtime):
            "Finite strain viscoelastic material formulation."

            # unimodular part of the right Cauchy-Green deformation tensor
            Cu = det(C) ** (-1 / 3) * C

            # update of state variables by evolution equation
            Ci = tm.special.from_triu_1d(Cin, like=C) + mu / eta * dtime * Cu
            Ci = det(Ci) ** (-1 / 3) * Ci

            # first invariant of elastic part of right Cauchy-Green deformation tensor
            I1 = tm.trace(Cu @ inv(Ci))

            # first Piola-Kirchhoff stress tensor and state variable
            return mu / 2 * (I1 - 3), tm.special.triu_1d(Ci)

        umat = fem.UserMaterialHyperelastic(
            viscoelastic, mu=1, eta=1, dtime=1, nstatevars=6
        )

    See the `documentation of tensortrax <https://github.com/adtzlr/tensortrax>`_
    for further details.

    """

    def __init__(self, fun, nstatevars=0, parallel=False, **kwargs):

        if nstatevars > 0:
            # split the original function into two sub-functions
            self.fun = tr.take(fun, item=0)
            self.fun_statevars = tr.take(fun, item=1)
        else:
            self.fun = fun

        self.parallel = parallel

        super().__init__(
            stress=self._stress,
            elasticity=self._elasticity,
            nstatevars=nstatevars,
            **kwargs
        )

    def _stress(self, x, **kwargs):
        F = x[0]

        if self.nstatevars > 0:
            statevars = (x[1],)
        else:
            statevars = ()

        C = dot(transpose(F), F)
        S = tr.gradient(self.fun, wrt=0, ntrax=2, parallel=self.parallel, sym=True)(
            C, *statevars, **kwargs
        )
        if self.nstatevars > 0:
            statevars_new = tr.function(
                self.fun_statevars, wrt=0, ntrax=2, parallel=self.parallel
            )(C, *statevars, **kwargs)
        else:
            statevars_new = None
        return [dot(F, 2 * S), statevars_new]

    def _elasticity(self, x, **kwargs):
        F = x[0]

        if self.nstatevars > 0:
            statevars = (x[1],)
        else:
            statevars = ()

        C = dot(transpose(F), F)
        D, S, W = tr.hessian(
            self.fun, wrt=0, ntrax=2, full_output=True, parallel=self.parallel, sym=True
        )(C, *statevars, **kwargs)
        A = np.einsum("iI...,kK...,IJKL...->iJkL...", F, F, 4 * D)
        B = cdya_ik(np.eye(3), 2 * S)
        return [np.sum(np.broadcast_arrays(A, B), axis=0)]



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import numpy as np
import pytest

import felupe as fe


def test_gausslegendre():
    q11 = fe.GaussLegendre(order=1, dim=1)
    assert q11.points.shape == (2, 1)
    assert q11.weights.sum() == 2

    for permute in [False, True]:
        q12 = fe.GaussLegendre(order=1, dim=2, permute=permute)
        assert q12.points.shape == (4, 2)
        assert np.isclose(q12.weights.sum(), 4)

        q13 = fe.GaussLegendre(order=1, dim=3, permute=permute)
        assert q13.points.shape == (8, 3)
        assert np.isclose(q13.weights.sum(), 8)

        q23 = fe.GaussLegendre(order=2, dim=3, permute=permute)
        assert q23.points.shape == (27, 3)
        assert np.isclose(q23.weights.sum(), 8)

    q23 = fe.GaussLegendre(order=2, dim=3)
    assert q23.inv().points.shape == q23.points.shape
    assert np.allclose(q23.weights, q23.inv().weights)

    with pytest.raises(ValueError):
        fe.GaussLegendre(order=1, dim=4)


def test_gausslegendre_boundary():

    with pytest.raises(ValueError):
        q11 = fe.GaussLegendreBoundary(order=1, dim=1)
        assert q11.points.shape == (2, 1)
        assert q11.weights.sum() == 2

    for permute in [False, True]:
        q12 = fe.GaussLegendreBoundary(order=1, dim=2, permute=permute)
        assert q12.points.shape == (2, 2)
        assert np.isclose(q12.weights.sum(), 2)

        q13 = fe.GaussLegendreBoundary(order=1, dim=3, permute=permute)
        assert q13.points.shape == (4, 3)
        assert np.isclose(q13.weights.sum(), 4)

        q23 = fe.GaussLegendreBoundary(order=2, dim=3, permute=permute)
        assert q23.points.shape == (9, 3)
        assert np.isclose(q23.weights.sum(), 4)

    q23 = fe.GaussLegendreBoundary(order=2, dim=3)
    assert q23.inv().points.shape == q23.points.shape
    assert np.allclose(q23.weights, q23.inv().weights)

    with pytest.raises(ValueError):
        fe.GaussLegendreBoundary(order=1, dim=4)


def test_triangle():
    q = fe.TriangleQuadrature(order=1)
    assert q.points.shape == (1, 2)
    assert q.weights.sum() == 1 / 2

    q = fe.TriangleQuadrature(order=2)
    assert q.points.shape == (3, 2)
    assert q.weights.sum() == 1 / 2

    q = fe.TriangleQuadrature(order=3)
    assert q.points.shape == (4, 2)
    assert q.weights.sum() == 1 / 2

    with pytest.raises(NotImplementedError):
        fe.TriangleQuadrature(order=4)


def test_tetra():
    q = fe.TetrahedronQuadrature(order=1)
    assert q.points.shape == (1, 3)
    assert q.weights.sum() == 1 / 6

    q = fe.TetrahedronQuadrature(order=2)
    assert q.points.shape == (4, 3)
    assert q.weights.sum() == 1 / 6

    q = fe.TetrahedronQuadrature(order=3)
    assert q.points.shape == (5, 3)
    assert q.weights.sum() == 1 / 6

    with pytest.raises(NotImplementedError):
        fe.TetrahedronQuadrature(order=4)


if __name__ == "__main__":
    test_gausslegendre()
    test_gausslegendre_boundary()
    test_triangle()
    test_tetra()



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 27 15:33:53 2021

@author: adutz
"""

import numpy as np


def test_readme():

    import felupe

    mesh = felupe.Cube(n=9)
    element = felupe.Hexahedron()
    quadrature = felupe.GaussLegendre(order=1, dim=3)

    region = felupe.Region(mesh, element, quadrature)

    dV = region.dV
    V = dV.sum()

    displacement = felupe.Field(region, dim=3)

    u = displacement.values
    ui = displacement.interpolate()
    dudX = displacement.grad()

    field = felupe.FieldContainer([displacement])

    F = field.extract(grad=True, sym=False, add_identity=True)

    umat = felupe.constitution.NeoHooke(mu=1.0, bulk=2.0)

    P = umat.gradient
    A = umat.hessian

    import numpy as np

    f0 = lambda x: np.isclose(x, 0)
    f1 = lambda x: np.isclose(x, 1)

    boundaries = {}
    boundaries["left"] = felupe.Boundary(displacement, fx=f0)
    boundaries["right"] = felupe.Boundary(displacement, fx=f1, skip=(1, 0, 0))
    boundaries["move"] = felupe.Boundary(displacement, fx=f1, skip=(0, 1, 1), value=0.5)

    dof0, dof1 = felupe.dof.partition(field, boundaries)
    ext0 = felupe.dof.apply(field, boundaries, dof0)

    linearform = felupe.IntegralForm(P(F)[:-1], field, dV)
    bilinearform = felupe.IntegralForm(A(F), field, dV, field)

    r = linearform.assemble().toarray()[:, 0]
    K = bilinearform.assemble()

    from scipy.sparse.linalg import spsolve  # default

    # from pypardiso import spsolve

    system = felupe.solve.partition(field, K, dof1, dof0, r)
    dfield = felupe.solve.solve(*system, ext0, solver=spsolve).reshape(*u.shape)

    # du = np.split(dfield, offsets)
    # field += du

    for iteration in range(8):
        F = field.extract()

        linearform = felupe.IntegralForm(P(F)[:-1], field, dV)
        bilinearform = felupe.IntegralForm(A(F), field, dV, field)

        r = linearform.assemble()
        K = bilinearform.assemble()

        system = felupe.solve.partition(field, K, dof1, dof0, r)
        dfield = felupe.solve.solve(*system, ext0, solver=spsolve).reshape(*u.shape)

        du = np.split(dfield, field.offsets)

        norm = felupe.math.norm(du)
        print(iteration, norm[0])
        field += du

        if iteration == 0:
            assert np.round(norm[0], 5) == 8.17418

        if norm[0] < 1e-12:
            break

    F[0][:, :, 0, 0]

    felupe.save(region, field, filename="result.vtk")

    from felupe.math import det, dot, transpose

    PK1 = P(F)[0]
    F = F[0]

    s = dot(PK1, transpose(F)) / det(F)

    # stress shifted and averaged to mesh-points
    cauchy_shifted = felupe.topoints(s, region, sym=True, mode="tensor")

    # stress projected and averaged to mesh-points
    cauchy_projected = felupe.project(s, region)

    felupe.save(
        region,
        field,
        filename="result_with_cauchy.vtk",
        point_data={
            "CauchyStressProjected": cauchy_projected,
            "CauchyStressShifted": cauchy_shifted,
        },
    )


if __name__ == "__main__":
    test_readme()



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
import pytest

import felupe as fe


def test_region():

    mesh = fe.Rectangle()
    r = fe.RegionQuad(mesh)
    r = fe.RegionQuadBoundary(mesh)
    r = fe.RegionQuadBoundary(mesh, ensure_3d=True)
    r = fe.RegionConstantQuad(mesh)

    mesh2 = fe.mesh.convert(mesh, 2, True, False, False)
    r = fe.RegionQuadraticQuad(mesh2)
    f = fe.FieldsMixed(r)

    r = fe.RegionQuadraticQuadBoundary(mesh2, ensure_3d=True)

    mesh3 = fe.mesh.convert(mesh, 2, True, True, False)
    r = fe.RegionQuad(mesh3)
    r = fe.RegionQuadraticQuad(mesh3)
    r = fe.RegionBiQuadraticQuad(mesh3)
    f = fe.FieldsMixed(r)

    r = fe.RegionBiQuadraticQuadBoundary(mesh3)

    mesh.cell_type = "some_fancy_cell_type"
    with pytest.raises(NotImplementedError):
        r = fe.RegionBoundary(mesh, fe.Quad(), fe.GaussLegendreBoundary(order=1, dim=2))

    mesh = fe.Cube()
    r = fe.RegionHexahedron(mesh)
    r = fe.RegionHexahedronBoundary(mesh)
    r = fe.RegionHexahedronBoundary(mesh, only_surface=False)
    r = fe.RegionHexahedronBoundary(mesh, mask=[0, 3])
    r = fe.RegionHexahedronBoundary(mesh, only_surface=False, mask=[0, 3])

    m = r.mesh_faces()
    assert np.allclose(m.points, mesh.points)
    assert m.cell_type == "quad"

    r = fe.RegionConstantHexahedron(mesh)

    mesh2 = fe.mesh.convert(mesh, 2, True, False, False)
    r = fe.RegionQuadraticHexahedron(mesh2)
    f = fe.FieldsMixed(r)

    r = fe.RegionQuadraticHexahedronBoundary(mesh2)

    mesh3 = fe.mesh.convert(mesh, 2, True, True, True)
    r = fe.RegionHexahedron(mesh3)
    r = fe.RegionQuadraticHexahedron(mesh3)
    r = fe.RegionTriQuadraticHexahedron(mesh3)
    f = fe.FieldsMixed(r)

    r = fe.RegionTriQuadraticHexahedronBoundary(mesh3)

    triangle = fe.Triangle()
    points = triangle.points
    cells = np.arange(3).reshape(1, -1)
    mesh = fe.Mesh(points, cells, "triangle")
    r = fe.RegionTriangle(mesh)

    mesh2 = fe.mesh.convert(mesh, 2, True, False, False)
    r = fe.RegionTriangle(mesh2)
    r = fe.RegionQuadraticTriangle(mesh2)
    f = fe.FieldsMixed(r)

    tetra = fe.Tetra()
    points = tetra.points
    cells = np.arange(4).reshape(1, -1)
    mesh = fe.Mesh(points, cells, "tetra")
    r = fe.RegionTetra(mesh)

    mesh2 = fe.mesh.convert(mesh, 2, True, False, False)
    r = fe.RegionTetra(mesh2)
    r = fe.RegionQuadraticTetra(mesh2)
    f = fe.FieldsMixed(r)

    triangle = fe.TriangleMINI()
    points = triangle.points
    cells = np.arange(4).reshape(1, -1)
    mesh = fe.Mesh(points, cells, "triangle-mini")
    r = fe.RegionTriangleMINI(mesh)
    f = fe.FieldsMixed(r)

    tetra = fe.TetraMINI()
    points = tetra.points
    cells = np.arange(5).reshape(1, -1)
    mesh = fe.Mesh(points, cells, "tetra-mini")
    r = fe.RegionTetraMINI(mesh)
    f = fe.FieldsMixed(r)

    order = 5
    dim = 3
    u = np.linspace(-1, 1, order + 1)
    z, y, x = np.meshgrid(u, u, u, indexing="ij")
    points = np.vstack((x.ravel(), y.ravel(), z.ravel())).T
    cells = np.arange(len(points)).reshape(1, -1)
    mesh = fe.Mesh(points, cells, "lagrange")
    r = fe.RegionLagrange(mesh, order, dim)


if __name__ == "__main__":
    test_region()



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import numpy as np
import pytest

import felupe as fe


def test_line2():
    line2 = fe.element.Line()

    r = [-1]

    h = line2.function(r)
    dhdr = line2.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -0.5)

    assert line2.shape == dhdr.shape


def test_quad0():
    quad0 = fe.element.ConstantQuad()

    r = [-1, -1]

    h = quad0.function(r)
    dhdr = quad0.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == 0)

    assert quad0.shape == dhdr.shape


def test_quad4():
    quad4 = fe.element.Quad()

    r = [-1, -1]

    h = quad4.function(r)
    dhdr = quad4.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -0.5)

    assert quad4.shape == dhdr.shape


def test_quad8():
    quad8 = fe.element.QuadraticQuad()

    r = [-1, -1]

    h = quad8.function(r)
    dhdr = quad8.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -1.5)

    assert quad8.shape == dhdr.shape


def test_quad9():
    quad9 = fe.element.BiQuadraticQuad()

    r = [-1, -1]

    h = quad9.function(r)
    dhdr = quad9.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -1.5)

    assert quad9.shape == dhdr.shape


def test_hex0():
    hex0 = fe.element.ConstantHexahedron()

    r = [-1, -1, -1]

    h = hex0.function(r)
    dhdr = hex0.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == 0)

    assert hex0.shape == dhdr.shape


def test_hex8():
    hex8 = fe.element.Hexahedron()

    r = [-1, -1, -1]

    h = hex8.function(r)
    dhdr = hex8.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -0.5)

    assert hex8.shape == dhdr.shape


def test_hex20():
    hex20 = fe.element.QuadraticHexahedron()

    r = [-1, -1, -1]

    h = hex20.function(r)
    dhdr = hex20.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -1.5)

    assert hex20.shape == dhdr.shape


def test_hex27():
    hex27 = fe.element.TriQuadraticHexahedron()

    r = [-1, -1, -1]

    h = hex27.function(r)
    dhdr = hex27.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -1.5)

    assert hex27.shape == dhdr.shape


def test_tri3():
    tri3 = fe.element.Triangle()

    r = [0, 0]

    h = tri3.function(r)
    dhdr = tri3.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -1)

    assert tri3.shape == dhdr.shape


def test_tri6():
    tri6 = fe.element.QuadraticTriangle()

    r = [0, 0]

    h = tri6.function(r)
    dhdr = tri6.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -3)

    assert tri6.shape == dhdr.shape


def test_tri_mini():
    trim = fe.element.TriangleMINI()

    r = [0, 0]

    h = trim.function(r)
    dhdr = trim.gradient(r)

    assert h[0] == 1
    assert h[-1] == 0  # check bubble
    assert np.all(dhdr[0] == -1)

    assert trim.shape == dhdr.shape


def test_tet4():
    tet4 = fe.element.Tetra()

    r = [0, 0, 0]

    h = tet4.function(r)
    dhdr = tet4.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -1)

    assert tet4.shape == dhdr.shape


def test_tet10():
    tet10 = fe.element.QuadraticTetra()

    r = [0, 0, 0]

    h = tet10.function(r)
    dhdr = tet10.gradient(r)

    assert h[0] == 1
    assert np.all(dhdr[0] == -3)

    assert tet10.shape == dhdr.shape


def test_tet_mini():
    tetm = fe.element.TetraMINI()

    r = [0, 0, 0]

    h = tetm.function(r)
    dhdr = tetm.gradient(r)

    assert h[0] == 1
    assert h[-1] == 0  # check bubble
    assert np.all(dhdr[0] == -1)

    assert tetm.shape == dhdr.shape


def test_aol():
    aol32 = fe.element.ArbitraryOrderLagrange(order=3, dim=2)
    aol23 = fe.element.ArbitraryOrderLagrange(order=2, dim=3)

    r = [-1, -1]

    h = aol32.function(r)
    dhdr = aol32.gradient(r)

    assert h[0] == 1
    assert aol32.shape == dhdr.shape

    r = [-1, -1, -1]

    h = aol23.function(r)
    dhdr = aol23.gradient(r)

    assert h[0] == 1
    assert aol23.shape == dhdr.shape


if __name__ == "__main__":
    test_line2()

    test_quad0()
    test_quad4()
    test_quad8()
    test_quad9()

    test_hex0()
    test_hex8()
    test_hex20()
    test_hex27()

    test_tri3()
    test_tri6()

    test_tri_mini()
    test_tet_mini()

    test_tet4()
    test_tet10()

    test_aol()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..math import sym as symmetric
from ._base import Field


class FieldPlaneStrain(Field):
    r"""A plane strain Field on points of a two-dimensional
    `region` with dimension `dim` (default is 2) and initial point
    `values` (default is 0).

    This is a modified :class:`Field` class in which the radial coordinates
    are evaluated at the numeric integration points. The :meth:`grad`-method is
    modified in such a way that it does not only contain the in-plane
    2d-gradient but also the circumferential stretch as shown in Eq.(1).

    ..  code-block::

                            |  dudX(2d) :   0   |
        dudX(planestrain) = | ..................|                  (1)
                            |     0     :   0   |

    Attributes
    ----------
    region : Region
        The region on which the field will be created.
    dim : int (default is 2)
        The dimension of the field.
    values : float (default is 0.0) or array
        A single value for all components of the field or an array of
        shape (region.mesh.npoints, dim)`.

    """

    def __init__(self, region, dim=2, values=0):
        """A plane strain Field on points of a two-dimensional
        `region` with dimension `dim` (default is 2) and initial point
        `values` (default is 0).

        Attributes
        ----------
        region : Region
            The region on which the field will be created.
        dim : int (default is 2)
            The dimension of the field.
        values : float (default is 0.0) or array
            A single value for all components of the field or an array of
            shape (region.mesh.npoints, dim)`.
        """

        # init base Field
        super().__init__(region, dim=dim, values=values)

    def _interpolate_2d(self):
        """Interpolate 2D field values at points and evaluate them at the
        integration points of all cells in the region."""

        # interpolated field values "aI"
        # evaluated at quadrature point "p"
        # for cell "c"
        return np.einsum(
            "ca...,apc->...pc", self.values[self.region.mesh.cells], self.region.h
        )

    def interpolate(self):
        # extend dimension of in-plane 2d-gradient
        return np.pad(self._interpolate_2d(), ((0, 1), (0, 0), (0, 0)))

    def _grad_2d(self, sym=False):
        """In-plane 2D gradient as partial derivative of field values at points
        w.r.t. the undeformed coordinates, evaluated at the integration points
        of all cells in the region. Optionally, the symmetric part of the
        gradient is returned.

        Arguments
        ---------
        sym : bool, optional (default is False)
            Calculate the symmetric part of the gradient.

        Returns
        -------
        array
            In-plane 2D-gradient as partial derivative of field values at points
            w.r.t. undeformed coordinates, evaluated at the integration points
            of all cells in the region.
        """

        # gradient as partial derivative of field component "I" at point "a"
        # w.r.t. undeformed coordinate "J" evaluated at quadrature point "p"
        # for each cell "e"
        g = np.einsum(
            "ca...,aJpc->...Jpc",
            self.values[self.region.mesh.cells],
            self.region.dhdX,
        )

        if sym:
            return symmetric(g)
        else:
            return g

    def grad(self, sym=False):
        """3D-gradient as partial derivative of field values at points w.r.t.
        the undeformed coordinates, evaluated at the integration points of all
        cells in the region. Optionally, the symmetric part of the gradient is
        returned.

        ..  code-block::

                                |  dudX(2d) :   0   |
            dudX(planestrain) = | ..................|
                                |     0     :   0   |

        Arguments
        ---------
        sym : bool, optional (default is False)
            Calculate the symmetric part of the gradient.

        Returns
        -------
        array
            Full 3D-gradient as partial derivative of field values at points
            w.r.t. undeformed coordinates, evaluated at the integration points
            of all cells in the region.
        """

        # extend dimension of in-plane 2d-gradient
        g = np.pad(self._grad_2d(sym=sym), ((0, 1), (0, 1), (0, 0), (0, 0)))

        return g



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.sparse import csr_matrix

from .._assembly import IntegralFormMixed
from ._helpers import Assemble, Results


class SolidBodyGravity:
    "A SolidBody with methods for the assembly of sparse vectors/matrices."

    def __init__(self, field, gravity=None, density=1.0):

        self.field = field
        self.results = Results(stress=False, elasticity=False)
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)
        self._form = IntegralFormMixed

        self.results.gravity = np.array(gravity)
        self.results.density = density

    def update(self, gravity):

        self.__init__(self.field, gravity, self.results.density)

    def _vector(self, field=None, parallel=False):

        if field is not None:
            self.field = field

        # copy and take only the first (displacement) field of the container
        f = self.field.copy()
        f.fields = f.fields[0:1]

        self.results.force = self._form(
            fun=[self.results.density * self.results.gravity.reshape(-1, 1, 1)],
            v=f,
            dV=self.field.region.dV,
            grad_v=[False],
        ).assemble(parallel=parallel)

        if len(self.field) > 1:
            self.results.force.resize(np.sum(self.field.fieldsizes), 1)

        return -self.results.force

    def _matrix(self, field=None, parallel=False):

        if field is not None:
            self.field = field

        n = np.sum(self.field.fieldsizes)
        self.results.stiffness = csr_matrix(([0], ([0], [0])), shape=(n, n))

        return self.results.stiffness



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import inspect
from time import perf_counter

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import spsolve

from .. import solve as fesolve
from .._assembly import IntegralFormMixed
from ..math import norm


class Result:
    def __init__(self, x, fun=None, jac=None, success=None, iterations=None):
        "Result class."
        self.x = x
        self.fun = fun
        self.jac = jac
        self.success = success
        self.iterations = iterations


def fun_items(items, x, parallel=False):
    "Force residuals from assembly of equilibrium (weak form)."

    # init keyword arguments
    kwargs = {"parallel": parallel}

    # link field of items with global field
    [item.field.link(x) for item in items]

    # init vector with shape from global field
    shape = (np.sum(x.fieldsizes), 1)
    vector = csr_matrix(shape)

    for body in items:
        # assemble vector
        r = body.assemble.vector(field=body.field, **kwargs)

        # check and reshape vector
        if r.shape != shape:
            r.resize(*shape)

        # add vector
        vector += r

    return vector.toarray()[:, 0]


def jac_items(items, x, parallel=False):
    "Tangent stiffness matrix from assembly of linearized equilibrium."

    # init keyword arguments
    kwargs = {"parallel": parallel}

    # init matrix with shape from global field
    shape = (np.sum(x.fieldsizes), np.sum(x.fieldsizes))
    matrix = csr_matrix(shape)

    for body in items:
        # assemble matrix
        K = body.assemble.matrix(**kwargs)

        # check and reshape matrix
        if K.shape != matrix.shape:
            K.resize(*shape)

        # add matrix
        matrix += K

    return matrix


def fun(x, umat, parallel=False, grad=True, add_identity=True, sym=False):
    "Force residuals from assembly of equilibrium (weak form)."

    return (
        IntegralFormMixed(
            fun=umat.gradient(x.extract(grad=grad, add_identity=add_identity, sym=sym))[
                :-1
            ],
            v=x,
            dV=x.region.dV,
        )
        .assemble(parallel=parallel)
        .toarray()[:, 0]
    )


def jac(x, umat, parallel=False, grad=True, add_identity=True, sym=False):
    "Tangent stiffness matrix from assembly of linearized equilibrium."

    return IntegralFormMixed(
        fun=umat.hessian(x.extract(grad=grad, add_identity=add_identity, sym=sym)),
        v=x,
        dV=x.region.dV,
        u=x,
    ).assemble(parallel=parallel)


def solve(A, b, x, dof1, dof0, offsets=None, ext0=None, solver=spsolve):
    "Solve partitioned system."

    system = fesolve.partition(x, A, dof1, dof0, -b)
    dx = fesolve.solve(*system, ext0, solver=solver)

    return dx


def check(dx, x, f, xtol, ftol, dof1=None, dof0=None, items=None, eps=1e-3):
    "Check result."

    def sumnorm(x):
        return np.sum(norm(x))

    xnorm = sumnorm(dx)

    if dof1 is None:
        dof1 = slice(None)

    if dof0 is None:
        dof0 = slice(0, 0)

    fnorm = sumnorm(f[dof1]) / (eps + sumnorm(f[dof0]))
    success = fnorm < ftol and xnorm < xtol

    if success and items is not None:
        for item in items:
            [item.results.update_statevars() for item in items]

    return xnorm, fnorm, success


def update(x, dx):
    "Update field."
    # x += dx # in-place
    return x + dx


def newtonrhapson(
    x0=None,
    fun=fun,
    jac=jac,
    solve=solve,
    maxiter=16,
    update=update,
    check=check,
    args=(),
    kwargs={},
    tol=np.sqrt(np.finfo(float).eps),
    items=None,
    dof1=None,
    dof0=None,
    ext0=None,
    solver=spsolve,
    verbose=True,
):
    """
    General-purpose Newton-Rhapson algorithm

    (Nonlinear) equilibrium equations `f`, as a function `f(x)` of the
    unknowns `x`, are solved by linearization of `f` at given unknowns `x0`.

        f(x0) = 0                                          (1)

        f(x0 + dx)     =  f(x0) + (df/dx)(x0) dx (= 0)     (2)
        (df/dx)(x0) dx = -f(x0)

        dx = solve(df/dx(x0), -f(x0))                      (3)

         x = x0 + dx                                       (4)

    Repeated evaluations of Eq.(3) and Eq.(4) lead to an incrementally updated
    solution of `x` which is shown in equation (4). Herein `xn` refer to the
    inital unknowns whereas `x` to the updated unknowns (the subscript `n+1`
    is dropped for readability).

        dx = solve(df/dx(xn), -f(xn))                      (5)

         x = xn + dx                                       (6)

    Eq.(5) and Eq.(6) are repeated until `check(dx, x, f)` returns `True`.

    """

    if verbose:
        runtimes = [perf_counter()]
        soltimes = []

    if x0 is not None:
        # take x0
        x = x0

    else:
        # obtain field of first body
        x = items[0].field

    kwargs_solve = {}
    sig = inspect.signature(solve)

    # pre-evaluate function at given unknowns "x"
    if items is not None:
        f = fun_items(items, x, *args, **kwargs)
    else:
        f = fun(x, *args, **kwargs)

    if verbose:
        print()
        print("Newton-Rhapson solver")
        print("=====================")
        print()
        print("| # | norm(fun) |  norm(dx) |")
        print("|---|-----------|-----------|")

    # iteration loop
    for iteration in range(maxiter):
        # evaluate jacobian at unknowns "x"
        if items is not None:
            K = jac_items(items, x, *args, **kwargs)
        else:
            K = jac(x, *args, **kwargs)

        # create keyword-arguments for solving the linear system
        keys = ["x", "dof1", "dof0", "ext0", "solver"]
        values = [x, dof1, dof0, ext0, solver]

        for key, value in zip(keys, values):
            if key in sig.parameters:
                kwargs_solve[key] = value

        # solve linear system and update solution
        if verbose:
            soltime_start = perf_counter()

        dx = solve(K, -f, **kwargs_solve)

        if verbose:
            soltime_end = perf_counter()
            soltimes.append([soltime_start, soltime_end])

        x = update(x, dx)

        # evaluate function at unknowns "x"
        if items is not None:
            f = fun_items(items, x, *args, **kwargs)
        else:
            f = fun(x, *args, **kwargs)

        # check success of solution
        xnorm, fnorm, success = check(
            dx=dx, x=x, f=f, xtol=np.inf, ftol=tol, dof1=dof1, dof0=dof0, items=items
        )

        if verbose:
            print("|%2d | %1.3e | %1.3e |" % (1 + iteration, fnorm, xnorm))

        if success:
            break

        if np.any(np.isnan([xnorm, fnorm])):
            raise ValueError("Norm of unknowns is NaN.")

    if 1 + iteration == maxiter and not success:
        raise ValueError("Maximum number of iterations reached (not converged).\n")

    Res = Result(x=x, fun=f, success=success, iterations=1 + iteration)

    if verbose:
        runtimes.append(perf_counter())
        runtime = np.diff(runtimes)
        soltime = np.diff(soltimes).sum()
        print(
            "\nConverged in %d iterations (Assembly: %1.4g s, Solve: %1.4g s).\n"
            % (iteration + 1, runtime - soltime, soltime)
        )

    return Res



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ..element import (
    ArbitraryOrderLagrange,
    BiQuadraticQuad,
    ConstantHexahedron,
    ConstantQuad,
    Hexahedron,
    Quad,
    QuadraticHexahedron,
    QuadraticQuad,
    QuadraticTetra,
    QuadraticTriangle,
    Tetra,
    TetraMINI,
    Triangle,
    TriangleMINI,
    TriQuadraticHexahedron,
)
from ..mesh import Mesh
from ..quadrature import GaussLegendre, GaussLegendreBoundary
from ..quadrature import Tetrahedron as TetraQuadrature
from ..quadrature import Triangle as TriangleQuadrature
from ._boundary import RegionBoundary
from ._region import Region


class RegionConstantQuad(Region):
    "A region with a constant quad element."

    def __init__(
        self,
        mesh,
        quadrature=GaussLegendre(order=1, dim=2),
        grad=False,
    ):

        element = ConstantQuad()
        super().__init__(mesh, element, quadrature, grad=grad)


class RegionQuad(Region):
    "A region with a quad element."

    def __init__(self, mesh, quadrature=GaussLegendre(order=1, dim=2), grad=True):

        element = Quad()

        if len(mesh.cells.T) > 4:
            mesh = Mesh(mesh.points, mesh.cells[:, :4], "quad")

        super().__init__(mesh, element, quadrature, grad=grad)


class RegionQuadraticQuad(Region):
    "A region with a (serendipity) quadratic quad element."

    def __init__(self, mesh, quadrature=GaussLegendre(order=2, dim=2), grad=True):

        element = QuadraticQuad()

        if len(mesh.cells.T) > 8:
            mesh = Mesh(mesh.points, mesh.cells[:, :8], "quad8")

        super().__init__(mesh, element, quadrature, grad=grad)


class RegionBiQuadraticQuad(Region):
    "A region with a bi-quadratic (lagrange) quad element."

    def __init__(self, mesh, quadrature=GaussLegendre(order=2, dim=2), grad=True):

        element = BiQuadraticQuad()

        super().__init__(mesh, element, quadrature, grad=grad)


class RegionQuadBoundary(RegionBoundary):
    "A region with a quad element."

    def __init__(
        self,
        mesh,
        quadrature=GaussLegendreBoundary(order=1, dim=2),
        grad=True,
        only_surface=True,
        mask=None,
        ensure_3d=False,
    ):

        element = Quad()

        super().__init__(
            mesh,
            element,
            quadrature,
            grad=grad,
            only_surface=only_surface,
            mask=mask,
            ensure_3d=ensure_3d,
        )


class RegionQuadraticQuadBoundary(RegionBoundary):
    "A boundary region with a quadratic quad element."

    def __init__(
        self,
        mesh,
        quadrature=GaussLegendreBoundary(order=2, dim=2),
        grad=True,
        only_surface=True,
        mask=None,
        ensure_3d=False,
    ):

        element = QuadraticQuad()

        super().__init__(
            mesh,
            element,
            quadrature,
            grad=grad,
            only_surface=only_surface,
            mask=mask,
            ensure_3d=ensure_3d,
        )


class RegionBiQuadraticQuadBoundary(RegionBoundary):
    "A boundary region with a bi-quadratic quad element."

    def __init__(
        self,
        mesh,
        quadrature=GaussLegendreBoundary(order=2, dim=2),
        grad=True,
        only_surface=True,
        mask=None,
        ensure_3d=False,
    ):

        element = BiQuadraticQuad()

        super().__init__(
            mesh,
            element,
            quadrature,
            grad=grad,
            only_surface=only_surface,
            mask=mask,
            ensure_3d=ensure_3d,
        )


class RegionConstantHexahedron(Region):
    "A region with a constant hexahedron element."

    def __init__(
        self,
        mesh,
        quadrature=GaussLegendre(order=1, dim=3),
        grad=False,
    ):

        element = ConstantHexahedron()
        super().__init__(mesh, element, quadrature, grad=grad)


class RegionHexahedron(Region):
    "A region with a hexahedron element."

    def __init__(self, mesh, quadrature=GaussLegendre(order=1, dim=3), grad=True):

        element = Hexahedron()

        if len(mesh.cells.T) > 8:
            mesh = Mesh(mesh.points, mesh.cells[:, :8], "hexahedron")

        super().__init__(mesh, element, quadrature, grad=grad)


class RegionHexahedronBoundary(RegionBoundary):
    "A region with a hexahedron element."

    def __init__(
        self,
        mesh,
        quadrature=GaussLegendreBoundary(order=1, dim=3),
        grad=True,
        only_surface=True,
        mask=None,
    ):

        element = Hexahedron()
        super().__init__(
            mesh, element, quadrature, grad=grad, only_surface=only_surface, mask=mask
        )


class RegionQuadraticHexahedron(Region):
    "A region with a (serendipity) quadratic hexahedron element."

    def __init__(self, mesh, quadrature=GaussLegendre(order=2, dim=3), grad=True):

        element = QuadraticHexahedron()

        if len(mesh.cells.T) > 20:
            mesh = Mesh(mesh.points, mesh.cells[:, :20], "hexahedron20")

        super().__init__(mesh, element, quadrature, grad=grad)


class RegionQuadraticHexahedronBoundary(RegionBoundary):
    "A boundary region with a (serendipity) quadratic hexahedron element."

    def __init__(
        self,
        mesh,
        quadrature=GaussLegendreBoundary(order=2, dim=3),
        grad=True,
        only_surface=True,
        mask=None,
    ):

        element = QuadraticHexahedron()
        super().__init__(
            mesh, element, quadrature, grad=grad, only_surface=only_surface, mask=mask
        )


class RegionTriQuadraticHexahedron(Region):
    "A region with a tri-quadratic (lagrange) hexahedron element."

    def __init__(self, mesh, quadrature=GaussLegendre(order=2, dim=3), grad=True):

        element = TriQuadraticHexahedron()
        super().__init__(mesh, element, quadrature, grad=grad)


class RegionTriQuadraticHexahedronBoundary(RegionBoundary):
    "A boundary region with a tri-quadratic (lagrange) hexahedron element."

    def __init__(
        self,
        mesh,
        quadrature=GaussLegendreBoundary(order=2, dim=3),
        grad=True,
        only_surface=True,
        mask=None,
    ):

        element = TriQuadraticHexahedron()
        super().__init__(
            mesh, element, quadrature, grad=grad, only_surface=only_surface, mask=mask
        )


class RegionLagrange(Region):
    "A region with an arbitrary order lagrange element."

    def __init__(self, mesh, order, dim, quadrature=None, grad=True):

        if quadrature is None:
            quadrature = GaussLegendre(order=order, dim=dim, permute=False)

        element = ArbitraryOrderLagrange(order, dim)
        self.order = order

        super().__init__(mesh, element, quadrature, grad=grad)


class RegionTriangle(Region):
    "A region with a triangle element."

    def __init__(self, mesh, quadrature=TriangleQuadrature(order=1), grad=True):

        element = Triangle()

        if len(mesh.cells.T) > 3:
            m = Mesh(mesh.points, mesh.cells[:, :3], "triangle")
        else:
            m = mesh

        super().__init__(m, element, quadrature, grad=grad)


class RegionTetra(Region):
    "A region with a tetra element."

    def __init__(self, mesh, quadrature=TetraQuadrature(order=1), grad=True):

        element = Tetra()

        if len(mesh.cells.T) > 4:
            m = Mesh(mesh.points, mesh.cells[:, :4], "tetra")
        else:
            m = mesh

        super().__init__(m, element, quadrature, grad=grad)


class RegionTriangleMINI(Region):
    "A region with a triangle-MINI element."

    def __init__(self, mesh, quadrature=TriangleQuadrature(order=2), grad=True):

        element = TriangleMINI()
        super().__init__(mesh, element, quadrature, grad=grad)


class RegionTetraMINI(Region):
    "A region with a tetra-MINI element."

    def __init__(self, mesh, quadrature=TetraQuadrature(order=2), grad=True):

        element = TetraMINI()
        super().__init__(mesh, element, quadrature, grad=grad)


class RegionQuadraticTriangle(Region):
    "A region with a quadratic triangle element."

    def __init__(self, mesh, quadrature=TriangleQuadrature(order=2), grad=True):

        element = QuadraticTriangle()
        super().__init__(mesh, element, quadrature, grad=grad)


class RegionQuadraticTetra(Region):
    "A region with a quadratic tetra element."

    def __init__(self, mesh, quadrature=TetraQuadrature(order=2), grad=True):

        element = QuadraticTetra()
        super().__init__(mesh, element, quadrature, grad=grad)



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import numpy as np
import pytest

import felupe as fe


def test_mpc():

    mesh = fe.Cube(n=3)
    mesh.points = np.vstack((mesh.points, [2, 0, 0]))
    mesh.update(mesh.cells)

    region = fe.RegionHexahedron(mesh)

    u = fe.FieldContainer([fe.Field(region, dim=3)])
    F = u.extract()

    umat = fe.constitution.NeoHooke(mu=1.0, bulk=2.0)

    f0 = lambda x: np.isclose(x, 0)
    f1 = lambda x: np.isclose(x, 1)
    f2 = lambda x: np.isclose(x, 2)

    boundaries = {}
    boundaries["left"] = fe.Boundary(u[0], fx=f0)
    boundaries["right"] = fe.Boundary(u[0], fx=f2, skip=(1, 0, 0))
    boundaries["move"] = fe.Boundary(u[0], fx=f2, skip=(0, 1, 1), value=0.5)

    mpc = fe.Boundary(u[0], fx=f1).points
    cpoint = mesh.npoints - 1

    RBE2 = fe.MultiPointConstraint(field=u, points=mpc, centerpoint=cpoint)

    for f in [None, u]:
        K_RBE2 = RBE2.assemble.matrix(u)
        r_RBE2 = RBE2.assemble.vector(u)

        linearform = fe.IntegralForm(umat.gradient(F)[:-1], u, region.dV)
        r = linearform.assemble() + r_RBE2

        bilinearform = fe.IntegralForm(umat.hessian(F), u, region.dV, u)
        K = bilinearform.assemble() + K_RBE2

        assert r.shape == (84, 1)
        assert K.shape == (84, 84)


def pre_mpc_mixed(point, values):

    mesh = fe.mesh.Cube(n=3)
    mesh.points = np.vstack((mesh.points, point))
    mesh.update(mesh.cells)

    region = fe.RegionHexahedron(mesh)
    dV = region.dV

    fields = fe.FieldsMixed(region, n=3)
    fields[0].values[-1] = values
    F, p, J = fields.extract()

    nh = fe.NeoHooke(mu=1.0, bulk=2.0)
    umat = fe.ThreeFieldVariation(nh)

    f0 = lambda x: np.isclose(x, 0)
    f1 = lambda x: np.isclose(x, 1)
    f2 = lambda x: np.isclose(x, 2)

    boundaries = {}
    boundaries["left"] = fe.Boundary(fields[0], fx=f0)
    boundaries["right"] = fe.Boundary(fields[0], fx=f2, skip=(1, 0, 0))
    boundaries["move"] = fe.Boundary(fields[0], fx=f2, skip=(0, 1, 1), value=0.5)

    mpc = fe.Boundary(fields[0], fx=f1).points
    cpoint = mesh.npoints - 1

    RBE2 = fe.MultiPointConstraint(fields, points=mpc, centerpoint=cpoint)
    CONT = fe.MultiPointContact(fields, points=mpc, centerpoint=cpoint)

    for f in [None, fields]:
        K_RBE2 = RBE2.assemble.matrix(f)
        r_RBE2 = RBE2.assemble.vector(f)

        K_CONT = CONT.assemble.matrix(f)
        r_CONT = CONT.assemble.vector(f)

        assert K_RBE2.shape == K_CONT.shape
        assert r_RBE2.shape == r_CONT.shape

    linearform = fe.IntegralForm(umat.gradient([F, p, J])[:-1], fields, dV)
    r = linearform.assemble()

    r_RBE2.resize(*r.shape)
    r = r + r_RBE2

    bilinearform = fe.IntegralForm(umat.hessian([F, p, J]), fields, dV, fields)
    K = bilinearform.assemble()

    K_RBE2.resize(*K.shape)
    K = K + K_RBE2


def test_mpc_mixed():
    pre_mpc_mixed(point=[2, 0, 0], values=[0, 0, 0])
    pre_mpc_mixed(point=[2, 0, 0], values=[-5, 0, 0])


def test_mpc_isolated():

    mesh = fe.mesh.Line(n=3)
    mesh.update(mesh.cells[:1])
    mesh.points = np.pad(mesh.points, ((0, 0), (0, 2)))
    mesh.points[-1] = np.array([1, 0.5, 0.5])
    mesh.dim = 3
    mesh.ndof = 9

    element = fe.Line()
    quadrature = fe.GaussLegendre(order=0, dim=1)
    region = fe.Region(mesh, element, quadrature, grad=False)
    field = fe.FieldContainer([fe.Field(region, dim=3)])

    # constraint

    field[0].values[-1] = [-0.6, 0, -0.6]

    # in x
    mpc = fe.MultiPointConstraint(
        field, points=[0, 1], centerpoint=2, skip=(0, 1, 1), multiplier=1e3
    )
    r = mpc.assemble.vector().toarray()
    K = mpc.assemble.matrix().toarray()

    assert np.allclose(r[[0, 3, 6]].ravel(), [600, 600, -1200])
    assert np.allclose(
        K[[0, 0, 3, 3, 6, 6, 6], [0, 6, 3, 6, 0, 3, 6]].ravel(),
        [1000, -1000, 1000, -1000, -1000, -1000, 2000],
    )

    # in z
    mpc = fe.MultiPointConstraint(
        field, points=[0, 1], centerpoint=2, skip=(1, 1, 0), multiplier=1e3
    )
    r = mpc.assemble.vector().toarray()
    K = mpc.assemble.matrix().toarray()

    assert np.allclose(r[[2, 5, 8]].ravel(), [600, 600, -1200])
    assert np.allclose(
        K[[2, 2, 5, 5, 8, 8, 8], [2, 8, 5, 8, 2, 5, 8]].ravel(),
        [1000, -1000, 1000, -1000, -1000, -1000, 2000],
    )

    # contact

    field[0].values[-1] = [-1.1, 0, -0.6]

    # in x
    mpc = fe.MultiPointContact(
        field, points=[0, 1], centerpoint=2, skip=(0, 1, 1), multiplier=1e3
    )
    r = mpc.assemble.vector().toarray()
    K = mpc.assemble.matrix().toarray()

    assert np.allclose(r[[0, 3, 6]].ravel(), [100, 600, -700])
    assert np.allclose(
        K[[0, 0, 3, 3, 6, 6, 6], [0, 6, 3, 6, 0, 3, 6]].ravel(),
        [1000, -1000, 1000, -1000, -1000, -1000, 2000],
    )

    # in z
    mpc = fe.MultiPointConstraint(
        field, points=[0, 1], centerpoint=2, skip=(1, 1, 0), multiplier=1e3
    )
    r = mpc.assemble.vector().toarray()
    K = mpc.assemble.matrix().toarray()

    assert np.allclose(r[[2, 5, 8]].ravel(), [600, 600, -1200])
    assert np.allclose(
        K[[2, 2, 5, 5, 8, 8, 8], [2, 8, 5, 8, 2, 5, 8]].ravel(),
        [1000, -1000, 1000, -1000, -1000, -1000, 2000],
    )

    # contact with partial active points

    field[0].values[-1] = [-0.6, 0, -0.6]

    # in x
    mpc = fe.MultiPointContact(
        field, points=[0, 1], centerpoint=2, skip=(0, 1, 1), multiplier=1e3
    )
    r = mpc.assemble.vector().toarray()
    K = mpc.assemble.matrix().toarray()

    assert np.allclose(r[[3, 6]].ravel(), [100, -100])
    assert np.allclose(
        K[[3, 3, 6, 6], [3, 6, 3, 6]].ravel(), [1000, -1000, -1000, 1000]
    )

    mesh.points[-2, 2] = -100

    # in z
    mpc = fe.MultiPointContact(
        field, points=[0, 1], centerpoint=2, skip=(1, 1, 0), multiplier=1e3
    )
    r = mpc.assemble.vector().toarray()
    K = mpc.assemble.matrix().toarray()

    assert np.allclose(r[[2, 8]].ravel(), [100, -100])
    assert np.allclose(
        K[[2, 2, 8, 8], [2, 8, 2, 8]].ravel(), [1000, -1000, -1000, 1000]
    )


if __name__ == "__main__":
    test_mpc()
    test_mpc_mixed()
    test_mpc_isolated()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

try:
    from einsumt import einsumt
except ModuleNotFoundError:
    from numpy import einsum as einsumt

from scipy.sparse import csr_matrix as sparsematrix


class IntegralForm:
    r"""Integral Form constructed by a function result ``fun``,
    a virtual field ``v``, differential volumes ``dV`` and optionally a
    field ``u``. For both fields ``v`` and ``u`` gradients may be passed by
    setting ``grad_v`` and ``grad_u`` to True (default is False for both).

    **Linearform**

    without gradient of ``v``

    ..  code-block::

        L(v) = ∫ fun v dV                                      (1)

           (or ∫ fun_i v_i dV)


    with gradient of ``v``

    ..  code-block::

        L(v) = ∫ fun grad(v) dV                                (2)

           (or ∫ fun_ij grad(v)_ij dV)


    **Bilinearform**

    without gradient of ``v`` and without gradient of ``u``

    ..  code-block::

        b(v, u) = ∫ v fun u dV                                 (3)

              (or ∫ v_i fun_ij u_j dV)

    with gradient of ``v`` and with gradient of ``u``

    ..  code-block::

        b(v, u) = ∫ grad(v) fun grad(u) dV                     (4)

              (or ∫ grad(v)_ij fun_ijkl grad(u)_kl dV)

    with gradient of ``v`` and without gradient of ``u``

    ..  code-block::

        b(v, u) = ∫ grad(v) fun u dV                           (5)

              (or ∫ grad(v)_ij fun_ijk u_k dV)

    without gradient of ``v`` and with gradient of ``u``

    ..  code-block::

        b(v, u) = ∫ v fun grad(u) dV                           (6)

              (or ∫ v_i fun_ikl grad(u)_kl dV)

    Arguments
    ---------
    fun : array
        The pre-evaluated function.
    v : Field
        The virtual Field.
    dV : array
        The differential volumes.
    u : Field, optional (default is None)
        If a Field is passed, a Bilinear-Form is created.
    grad_v : bool, optional (default is False)
        Flag to activate the gradient on Field ``v``.
    grad_u : bool, optional (default is False)
        Flag to activate the gradient on Field ``u``.
    """

    def __init__(self, fun, v, dV, u=None, grad_v=False, grad_u=False):
        self.fun = np.ascontiguousarray(fun)
        self.dV = dV

        self.v = v
        self.grad_v = grad_v

        self.u = u
        self.grad_u = grad_u

        # init indices

        # # linear form
        if not self.u:
            self.indices = self.v.indices.ai
            self.shape = self.v.indices.shape

        # # bilinear form
        else:
            eai = self.v.indices.eai
            ebk = self.u.indices.eai

            eaibk0 = np.repeat(eai, ebk.shape[1] * self.u.dim)
            eaibk1 = np.tile(ebk, (1, eai.shape[1] * self.v.dim, 1)).ravel()

            self.indices = (eaibk0, eaibk1)
            self.shape = (self.v.indices.shape[0], self.u.indices.shape[0])

    def assemble(self, values=None, parallel=False):
        "Assembly of sparse region vectors or matrices."

        if values is None:
            values = self.integrate(parallel=parallel)

        permute = np.append(len(values.shape) - 1, range(len(values.shape) - 1)).astype(
            int
        )

        out = sparsematrix(
            (values.transpose(permute).ravel(), self.indices), shape=self.shape
        )

        return out

    def integrate(self, parallel=False):
        "Return evaluated (but not assembled) integrals."

        grad_v, grad_u = self.grad_v, self.grad_u
        v, u = self.v, self.u
        dV = self.dV
        fun = self.fun

        # plane strain
        # trim 3d vector-valued functions to the dimension of the field
        function_dimension = len(fun.shape) - 2
        function_is_vector = function_dimension >= 1
        function_is_3d = len(fun) == 3
        field_is_2d = v.dim == 2

        if function_is_vector and function_is_3d and field_is_2d:
            fun = fun[tuple([slice(2)] * function_dimension)]

        if parallel:
            einsum = einsumt
        else:
            einsum = np.einsum

        if not grad_v:
            vb = v.region.h
        else:
            vb = v.region.dhdX

        if u is not None:
            if not grad_u:
                ub = u.region.h
            else:
                ub = u.region.dhdX

        if u is None:

            if not grad_v:
                return einsum("ape,...pe,pe->a...e", vb, fun, dV, optimize=True)
            else:
                return einsum("aJpe,...Jpe,pe->a...e", vb, fun, dV, optimize=True)

        else:

            if not grad_v and not grad_u:
                out = einsum("ape,...pe,bpe,pe->a...be", vb, fun, ub, dV, optimize=True)
                if len(out.shape) == 5:
                    return einsum("aijbe->aibje", out)
                else:
                    return out
            elif grad_v and not grad_u:
                return einsum(
                    "aJpe,iJ...pe,bpe,pe->aib...e", vb, fun, ub, dV, optimize=True
                )
            elif not grad_v and grad_u:
                return einsum(
                    "a...pe,...kLpe,bLpe,pe->a...bke",
                    vb,
                    fun,
                    ub,
                    dV,
                    optimize=True,
                )
            else:  # grad_v and grad_u
                return einsum(
                    "aJpe,iJkLpe,bLpe,pe->aibke", vb, fun, ub, dV, optimize=True
                )



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from .._assembly import IntegralFormMixed
from ..constitution import AreaChange
from ..math import det, dot, transpose
from ._helpers import Assemble, Evaluate, Results


class SolidBody:
    "A SolidBody with methods for the assembly of sparse vectors/matrices."

    def __init__(self, umat, field, statevars=None):

        self.umat = umat
        self.field = field

        self.results = Results(stress=True, elasticity=True)
        self.results.kinematics = self._extract(self.field)

        if statevars is not None:
            self.results.statevars = statevars
        else:
            self.results.statevars = np.zeros(
                (
                    *umat.x[-1].shape,
                    field.region.quadrature.npoints,
                    field.region.mesh.ncells,
                )
            )

        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)

        self.evaluate = Evaluate(
            gradient=self._gradient,
            hessian=self._hessian,
            cauchy_stress=self._cauchy_stress,
            kirchhoff_stress=self._kirchhoff_stress,
        )

        self._area_change = AreaChange()

        self._form = IntegralFormMixed

    def _vector(self, field=None, parallel=False, items=None, args=(), kwargs={}):

        if field is not None:
            self.field = field

        self.results.stress = self._gradient(field, args=args, kwargs=kwargs)
        self.results.force = self._form(
            fun=self.results.stress[slice(items)],
            v=self.field,
            dV=self.field.region.dV,
        ).assemble(parallel=parallel)

        return self.results.force

    def _matrix(self, field=None, parallel=False, items=None, args=(), kwargs={}):

        if field is not None:
            self.field = field

        self.results.elasticity = self._hessian(field, args=args, kwargs=kwargs)

        self.results.stiffness = self._form(
            fun=self.results.elasticity[slice(items)],
            v=self.field,
            u=self.field,
            dV=self.field.region.dV,
        ).assemble(parallel=parallel)

        return self.results.stiffness

    def _extract(self, field):

        self.field = field
        self.results.kinematics = self.field.extract()

        return self.results.kinematics

    def _gradient(self, field=None, args=(), kwargs={}):

        if field is not None:
            self.field = field
            self.results.kinematics = self._extract(self.field)

        gradient = self.umat.gradient(
            [*self.results.kinematics, self.results.statevars], *args, **kwargs
        )

        self.results.stress, self.results._statevars = gradient[:-1], gradient[-1]

        return self.results.stress

    def _hessian(self, field=None, args=(), kwargs={}):

        if field is not None:
            self.field = field
            self.results.kinematics = self._extract(self.field)

        self.results.elasticity = self.umat.hessian(
            [*self.results.kinematics, self.results.statevars], *args, **kwargs
        )

        return self.results.elasticity

    def _kirchhoff_stress(self, field=None):

        self._gradient(field)

        P = self.results.stress[0]
        F = self.results.kinematics[0]

        return dot(P, transpose(F))

    def _cauchy_stress(self, field=None):

        self._gradient(field)

        P = self.results.stress[0]
        F = self.results.kinematics[0]
        J = det(F)

        return dot(P, transpose(F)) / J



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ..region._templates import (
    RegionBiQuadraticQuad,
    RegionConstantHexahedron,
    RegionConstantQuad,
    RegionHexahedron,
    RegionLagrange,
    RegionQuad,
    RegionQuadraticHexahedron,
    RegionQuadraticQuad,
    RegionQuadraticTetra,
    RegionQuadraticTriangle,
    RegionTetra,
    RegionTetraMINI,
    RegionTriangle,
    RegionTriangleMINI,
    RegionTriQuadraticHexahedron,
)
from ._axi import FieldAxisymmetric
from ._base import Field
from ._container import FieldContainer
from ._planestrain import FieldPlaneStrain


class FieldsMixed(FieldContainer):
    r"""A mixed field based on a region and returns a :class:`FieldContainer`
    instance."""

    def __init__(
        self,
        region,
        n=3,
        values=(0, 0, 1, 0),
        axisymmetric=False,
        planestrain=False,
        offset=0,
        npoints=None,
        mesh=None,
        **kwargs,
    ):
        r"""Create a mixed field based on a region. The dual region is chosen
        automatically, i.e. for a :class:`RegionHexahedron` the dual region
        is :class:`RegionConstantHexahedron`. A total number of ``n`` fields
        are generated and passed to :class:`FieldContainer`.

        Arguments
        ---------
        region : Region
            A template region.
        n : int, optional (default is 2)
            Number of fields where the first one is a vector field of mesh-
            dimension and the following fields are scalar-fields.
        values : tuple, optional (default is (0, 0, 1, 0))
            Initial field values.
        axisymmetric : bool, optional
            Flag to initiate a axisymmetric Field (default is False).
        planestrain : bool, optional
            Flag to initiate a plane strain Field (default is False).
        offset : int, optional
            Offset for cell connectivity (default is 0).
        npoints : int or None, optional
            Specified number of mesh points (default is None).
        mesh: Mesh or None, optional
            A mesh for the dual region (default is None).
        """

        regions = {
            RegionHexahedron: RegionConstantHexahedron,
            RegionQuad: RegionConstantQuad,
            RegionQuadraticQuad: RegionConstantQuad,
            RegionBiQuadraticQuad: RegionQuad,
            RegionQuadraticHexahedron: RegionConstantHexahedron,
            RegionTriQuadraticHexahedron: RegionHexahedron,
            RegionQuadraticTetra: RegionTetra,
            RegionQuadraticTriangle: RegionTriangle,
            RegionTetraMINI: RegionTetra,
            RegionTriangleMINI: RegionTriangle,
            RegionLagrange: RegionLagrange,
        }
        points_per_cell = {
            RegionConstantHexahedron: 1,
            RegionConstantQuad: 1,
            RegionQuad: 4,
            RegionHexahedron: 8,
            RegionTriangle: 3,
            RegionTetra: 4,
            RegionLagrange: None,
        }

        kwargs0 = {}

        if isinstance(region, RegionLagrange):
            kwargs0["order"] = region.order - 1
            kwargs0["dim"] = region.mesh.dim
            points_per_cell[RegionLagrange] = region.order**region.element.dim

        RegionDual = regions[type(region)]

        if mesh is None and points_per_cell[RegionDual] is not None:
            mesh = region.mesh.dual(
                points_per_cell=points_per_cell[RegionDual],
                offset=offset,
                npoints=npoints,
            )

        region_dual = RegionDual(
            mesh,
            quadrature=region.quadrature,
            grad=False,
            **{**kwargs0, **kwargs},
        )

        if axisymmetric is False and planestrain is False:
            F = Field
        elif axisymmetric is True and planestrain is False:
            F = FieldAxisymmetric
        elif axisymmetric is False and planestrain is True:
            F = FieldPlaneStrain

        fields = [F(region, dim=region.mesh.dim, values=values[0])]

        for a in range(1, n):
            fields.append(Field(region_dual, values=values[a]))

        super().__init__(fields=tuple(fields))



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._container import MeshContainer
from ._mesh import Mesh


def read(
    filename, file_format=None, cellblock=None, dim=None, merge=False, decimals=None
):
    "Read a mesh-file using meshio and create a felupe mesh-container."

    from meshio import read as meshio_read

    m = meshio_read(filename=filename, file_format=file_format)

    if dim is None:
        dim = m.points.shape[1]

    points = m.points[:, :dim]

    if cellblock is None:
        cellblock = slice(None)

    cells = m.cells[cellblock]

    if not isinstance(cells, list):
        cells = [cells]

    if len(cells) > 0:
        meshes = [Mesh(points, c.data, c.type) for c in cells]
    else:
        meshes = [Mesh(points, np.zeros((0, 0), dtype=int), None)]

    return MeshContainer(meshes, merge=merge, decimals=decimals)



from ._base import Scheme
from ._gausslegendre import GaussLegendre, GaussLegendreBoundary
from ._tetra import Tetrahedron
from ._triangle import Triangle

__all__ = [
    "Scheme",
    "GaussLegendre",
    "GaussLegendreBoundary",
    "Tetrahedron",
    "Triangle",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._tools import expand


def line_line(a=0, b=1, n=2):
    "Line generator."
    points = np.linspace(a, b, n).reshape(-1, 1)
    cells = np.repeat(np.arange(n), 2)[1:-1].reshape(-1, 2)
    cell_type = "line"

    return points, cells, cell_type


def rectangle_quad(a=(0, 0), b=(1, 1), n=(2, 2)):
    "Rectangle generator."
    dim = 2
    array_like = (tuple, list, np.ndarray)

    # check if number "n" is scalar or no. of points per axis (array-like)
    if not isinstance(n, array_like):
        n = np.full(dim, n, dtype=int)

    line = line_line(a=a[0], b=b[0], n=n[0])

    points, cells, cell_type = expand(*line, n=n[-1], z=b[-1] - a[-1])
    points[:, -1] += a[-1]

    return points, cells, cell_type


def cube_hexa(a=(0, 0, 0), b=(1, 1, 1), n=(2, 2, 2)):
    "Cube generator."
    dim = 3
    array_like = (tuple, list, np.ndarray)

    # check if number "n" is scalar or no. of points per axis (array-like)
    if not isinstance(n, array_like):
        n = np.full(dim, n, dtype=int)

    rectangle = rectangle_quad(a=a[:-1], b=b[:-1], n=n[:-1])

    points, cells, cell_type = expand(*rectangle, n=n[-1], z=b[-1] - a[-1])
    points[:, -1] += a[-1]

    return points, cells, cell_type



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


class Boundary:
    "A Boundary as a collection of prescribed degrees of freedom."

    def __init__(
        self,
        field,
        name="default",
        fx=lambda v: v == np.nan,
        fy=lambda v: v == np.nan,
        fz=lambda v: v == np.nan,
        value=0,
        skip=(False, False, False),
        mask=None,
        mode="or",
    ):
        """A Boundary as a collection of prescribed degrees of freedom
        (numbered coordinate components of a field at points of a mesh).


        Arguments
        ---------
        field : Field
            Field on wich the boundary is created.

        name : str, optional (default is "default")
            Name of the boundary.

        fx : function, optional (default is `lambda v: v == np.nan`)
            Mask-function for x-component of mesh-points which returns
            `True` at points on which the boundary will be applied.

        fy : function, optional (default is `lambda v: v == np.nan`)
            Mask-function for y-component of mesh-points which returns
            `True` at points on which the boundary will be applied.

        fz : function, optional (default is `lambda v: v == np.nan`)
            Mask-function for z-component of mesh-points which returns
            `True` at points on which the boundary will be applied.

        value : int, optional (default is 0)
            Value of the selected (prescribed) degrees of freedom.

        skip : tuple of bool, optional (default is `(False, False, False)`)
            A tuple to define which axes of the selected points should be
            skipped (i.e. not prescribed).

        mode : string, optional (default is `or`)
            A string which defines the logical combination of points per axis.


        Attributes
        ----------
        mask : array
            Mask-array which contains prescribed degrees of freedom.

        dof :array
            Array which contains prescribed degrees of freedom.

        points : array
            Array which contains the points on which one or more degrees of
            freedom are prescribed.

        """

        mesh = field.region.mesh
        dof = field.indices.dof

        self.field = field
        self.dim = field.dim  # mesh.dim
        self.name = name
        self.value = value
        self.skip = np.array(skip).astype(int)[: mesh.dim]  # self.dim

        self.mode = mode

        # check if callable
        _fx = fx if callable(fx) else lambda x: np.isclose(x, fx)
        _fy = fy if callable(fy) else lambda y: np.isclose(y, fy)
        _fz = fz if callable(fz) else lambda z: np.isclose(z, fz)

        self.fun = [_fx, _fy, _fz][: mesh.dim]

        if mask is None:

            # apply functions on the points per coordinate
            # fx(x), fy(y), fz(z) and create a mask for each coordinate
            mask = [f(x) for f, x in zip(self.fun, mesh.points.T)]

            # select the logical combination function "or" or "and"
            combine = {"or": np.logical_or, "and": np.logical_and}[self.mode]

            # combine the masks with "logical_or" if dim > 1
            if mesh.dim == 1:
                mask = mask[0]

            elif mesh.dim == 2:
                mask = combine(mask[0], mask[1])

            elif mesh.dim == 3:  # and mesh.points.shape[1] == 3:
                tmp = np.logical_or(mask[0], mask[1])
                mask = combine(tmp, mask[2])

        # tile the mask
        self.mask = np.tile(mask.reshape(-1, 1), self.dim)

        # check if some axes should be skipped
        if True not in skip:
            pass
        else:
            # exclude mask from axes which should be skipped
            self.mask[:, np.where(self.skip)[0]] = False

        self.dof = dof[self.mask]
        self.points = np.arange(mesh.npoints)[mask]

    def update(self, value):

        self.value = value



from ._axi import FieldAxisymmetric
from ._base import Field
from ._container import FieldContainer
from ._fields import FieldsMixed
from ._planestrain import FieldPlaneStrain

__all__ = [
    "FieldAxisymmetric",
    "Field",
    "FieldContainer",
    "FieldsMixed",
    "FieldPlaneStrain",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


def rotation_matrix(alpha_deg, dim=3, axis=0):
    """Rotation matrix with given rotation axis and dimension (2d or 3d).

    Parameters
    ----------
    alpha_deg : int
        Rotation angle in degree.
    dim : int, optional (default is 3)
        Dimension of the rotation matrix.
    axis : int, optional (default is 0)
        Rotation axis.

    Returns
    -------
    rotation_matrix : ndarray
        Rotation matrix of dim 2 or 3 with given rotation axis.
    """

    a = np.deg2rad(alpha_deg)
    rotation_matrix = np.array([[np.cos(a), -np.sin(a)], [np.sin(a), np.cos(a)]])

    if dim == 3:
        if axis == 1:
            rotation_matrix = rotation_matrix.T
        rotation_matrix = np.insert(rotation_matrix, [axis], np.zeros((1, 2)), axis=0)
        rotation_matrix = np.insert(rotation_matrix, [axis], np.zeros((3, 1)), axis=1)
        rotation_matrix[axis, axis] = 1

    return rotation_matrix



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..tools import force
from ._job import Job


class CharacteristicCurve(Job):
    def __init__(
        self,
        steps,
        boundary,
        items=None,
        callback=lambda stepnumber, substepnumber, substep: None,
    ):

        super().__init__(steps, self._callback)

        self.items = items
        self.boundary = boundary
        self.x = []
        self.y = []
        self.res = None
        self._cb = callback

    def _callback(self, stepnumber, substepnumber, substep):

        if self.items is not None:
            fun = sum([item.results.force for item in self.items])
        else:
            fun = substep.fun

        self.x.append(substep.x[0].values[self.boundary.points[0]])
        self.y.append(force(substep.x, fun, self.boundary))
        self.res = substep

        self._cb(stepnumber, substepnumber, substep)

    def plot(
        self,
        x=None,
        y=None,
        xaxis=0,
        yaxis=0,
        xlabel="x",
        ylabel="y",
        xscale=1,
        yscale=1,
        gradient=False,
        fig=None,
        ax=None,
        items=None,
        **kwargs
    ):
        if self.res is None:
            raise ValueError(
                "Results are empty. Run `job.evaluate()` and call `job.plot()` again."
            )

        import matplotlib.pyplot as plt

        if x is None:
            x = self.x
        if y is None:
            y = self.y

        if items is None:
            items = slice(None)

        x = np.array(x)[items]
        y = np.array(y)[items]

        if gradient:
            y = np.gradient(y, x[:, xaxis], edge_order=2, axis=0)
            z = np.gradient(y, x[:, xaxis], edge_order=2, axis=0)
            cuttoff = np.mean(abs(z[:, yaxis])) + 2 * np.std(abs(z[:, yaxis]))
            y[abs(z) > cuttoff] = np.nan

        if fig is None or ax is None:
            fig, ax = plt.subplots()

        ax.plot(x[:, xaxis] * xscale, y[:, yaxis] * yscale, **kwargs)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        return fig, ax



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np

from .._assembly import IntegralFormMixed
from .._field import FieldAxisymmetric
from ..constitution import AreaChange
from ..math import det


class Assemble:
    "A class with assembly methods of a SolidBody."

    def __init__(self, vector, matrix):
        self.vector = vector
        self.matrix = matrix


class Evaluate:
    "A class with evaluate methods of a SolidBody."

    def __init__(self, gradient, hessian, cauchy_stress=None, kirchhoff_stress=None):
        self.gradient = gradient
        self.hessian = hessian

        if cauchy_stress is not None:
            self.cauchy_stress = cauchy_stress
            self.kirchhoff_stress = kirchhoff_stress


class Results:
    "A class with intermediate results of a SolidBody."

    def __init__(self, stress=False, elasticity=False):

        self.force = None
        self._force = None
        self.stiffness = None
        self.kinematics = None
        self.statevars = None
        self._statevars = None

        if stress:
            self.stress = None

        if elasticity:
            self.elasticity = None

    def update_statevars(self):

        if self._statevars is not None:
            self.statevars = self._statevars


class StateNearlyIncompressible:
    "A State with internal fields for (nearly) incompressible solid bodies."

    def __init__(self, field):

        self.field = field
        self.dJdF = AreaChange().function

        # initial values (on mesh-points) of the displacement field
        self.u = field[0].values

        # deformation gradient
        self.F = field.extract()

        # cell-values of the internal pressure and volume-ratio fields
        self.p = np.zeros(field.region.mesh.ncells)
        self.J = np.ones(field.region.mesh.ncells)

    def h(self, parallel=False):
        "Integrated shape-function gradient w.r.t. the deformed coordinates `x`."

        return IntegralFormMixed(
            fun=self.dJdF(self.F), v=self.field, dV=self.field.region.dV
        ).integrate(parallel=parallel)[0]

    def v(self):
        "Cell volumes of the deformed configuration."
        dV = self.field.region.dV
        if isinstance(self.field[0], FieldAxisymmetric):
            R = self.field[0].radius
            dA = self.field.region.dV
            dV = 2 * np.pi * R * dA
        return (det(self.F[0]) * dV).sum(0)
